"""This module defines the WeclappMetaData class, which represents custom attributes
in Weclapp entities. It includes methods for setting, adding, and removing values,
validating values, and checking if a value is present."""

import logging
from datetime import datetime
from typing import Union, Any, Literal
from pydantic import BaseModel
from ...time_functions import convert_to_time_format
from . import config


class WeclappMetaData(BaseModel):
    """Class representing custom attributes in Weclapp entities."""

    attributeDefinitionId: str
    valueName: Union[config.CUSTOM_ATTRIBUTE_TYPES_LITERAL, None] = None
    value: Any = None
    originalValue: Any = None
    empty: bool = False

    def __init__(self, **data):
        attribute_definition_id = data["attributeDefinitionId"]
        value_name = None
        value = (
            []
            if any(key in data for key in config.LIST_CUSTOM_ATTRIBUTE_TYPES)
            else None
        )
        for key in config.CUSTOM_ATTRIBUTE_TYPES:
            if key in data:
                value_name = key
                value = data.get(key)
                break

        super().__init__(
            attributeDefinitionId=attribute_definition_id,
            valueName=value_name,
            value=value,
            originalValue=value,
        )

    @property
    def updated(self) -> bool:
        """Checks if the value was updated and differs from the original value."""
        if self.originalValue == self.value:
            return False
        if self.originalValue is None:
            return True
        if self.value is not None:
            if self.valueName in config.FLOAT_CUSTOM_ATTRIBUTE_TYPES:
                if float(self.originalValue) == float(self.value):
                    return False
            if self.valueName in config.INT_CUSTOM_ATTRIBUTE_TYPES:
                if int(self.originalValue) == int(self.value):
                    return False
            if self.valueName in config.LIST_CUSTOM_ATTRIBUTE_TYPES:
                if sorted(self.originalValue, key=lambda x: str(x)) == sorted(
                    self.value, key=lambda x: str(x)
                ):
                    return False
        return True

    def set_value(
        self,
        value: Any,
        value_name: config.CUSTOM_ATTRIBUTE_TYPES_LITERAL,
        empty_field: bool = False,
    ) -> None:
        """Method to set a new value to a custom attribute. If empty_field is True,
        the update dictionary will not contain field value and the custom attribute
        field will be empty in Weclapp after update."""
        self.valueName = value_name

        if self.valueName not in config.SINGLE_VALUE_CUSTOM_ATTRIBUTE_TYPES:
            raise AssertionError(
                "Function set_value() is only for single value custom attributes, "
                f"not for {self.valueName}"
            )

        if empty_field:
            self.empty = True
            self.value = None

        else:
            self.value = self.validate_value(value)

    def add_value(self, value: Any, value_name: config.CUSTOM_ATTRIBUTE_TYPES_LITERAL) -> None:
        """Add a value to a list of values. This function is only for
        custom attributes that have multiselection capabilities, such as
        selectedValues and entityReferences."""
        self.valueName = value_name

        if self.valueName not in config.LIST_CUSTOM_ATTRIBUTE_TYPES:
            raise AssertionError(
                "Function add_value() is only for list custom attributes, "
                f"not for {self.valueName}"
            )

        if self.valueName == "selectedValues":
            selected_value_ids = {element["id"] for element in self.value}
            selected_value_ids.add(value)
            updated_values_list = [{"id": el} for el in selected_value_ids]
        else:
            if not ("entityId" in value and "entityName" in value):
                raise AssertionError(
                    "Value of entityReferences need to have entityId and entityName"
                )
            if not value["entityId"] in [entity["entityId"] for entity in self.value]:
                updated_values_list = self.value
                updated_values_list.append(value)
            else:
                updated_values_list = self.value

        self.value = self.validate_value(updated_values_list)

    def remove_value(self, value: Any, value_name: config.CUSTOM_ATTRIBUTE_TYPES_LITERAL) -> None:
        """Removes a value from a list of values. This function is only for
        custom attributes that have multiselection capabilities, such as
        selectedValues and entityReferences."""
        self.valueName = value_name

        if self.valueName not in config.LIST_CUSTOM_ATTRIBUTE_TYPES:
            raise AssertionError(
                "Function add_value() is only for list custom attributes, "
                f"not for {self.valueName}"
            )

        updated_values_list = []
        key = "id" if self.valueName == "selectedValues" else "entityId"
        for element in self.value:
            if element[key] != value:
                updated_values_list.append(element)

        self.value = self.validate_value(updated_values_list)

    def has_value(self, target_value: Any) -> bool:
        """Checks for selectedValues and entityReferences if a targetValue is
        present. Here valueName will always be provided, because Weclapp API
        always shows multiselection values as lists, even if it is empty. If not
        provided, it is then assumed that this is not a multiselection
        custom attribute."""
        if target_value is None:
            logging.error("target_value is None, cannot check for presence")
            return False
        if self.valueName is None:
            logging.error(
                "valueName is None, check if this is a custom attribute of "
                "multiselection type"
            )
            return False

        if self.valueName == "selectedValues":
            return target_value in [
                el["id"] for el in self.value
            ]
        if self.valueName == "entityReferences":
            return target_value in [
                el["entityId"] for el in self.value
            ]
        return self.value == target_value

    def validate_value(self, value: Any) -> Any:
        """Validates the custom attribute value and converts it to the correct
        format. Raises an error if the value format is invalid and cannot be converted.
        """

        if self.valueName == "numberValue":
            if type(value) in [int, float]:
                return str(round(value, 2))
            if not str(value).replace(".", "").strip("-").isnumeric():
                raise ValueError(
                    f"Given value ({value}) is not numeric. It needs to be "
                    "a float or int, or a string that can be converted to a float."
                )
            return str(value)

        if self.valueName == "booleanValue":
            if not isinstance(value, bool):
                raise ValueError(
                    f"Given value ({value}) is not boolean. It needs to be a boolean."
                )
            return bool(value)

        if self.valueName in ["stringValue", "entityId", "selectedValueId"]:
            return str(value)

        if self.valueName == "dateValue":
            if isinstance(value, datetime):
                value = convert_to_time_format(value, conversion_format="weclapp")
            if not isinstance(value, int):
                raise ValueError(
                    f"Date value ({value}) is not int. It needs to be an int "
                    " or a datetime object."
                )
            return int(value)

        if self.valueName in config.LIST_CUSTOM_ATTRIBUTE_TYPES:
            if not isinstance(value, list):
                raise ValueError(
                    f"Value ({value}) must be a list, but is {type(value).__name__}"
                )
            for element in value:
                if not isinstance(element, dict):
                    raise ValueError(
                        f"All elements in the list must be dicts, but "
                        f"found {type(element).__name__}"
                    )
            return value

        raise ValueError("Error in cAtt.setValue(): Something is seriously wrong")

    def build_update_dictionary(
        self,
        update_type: Literal["full", "used"] = "used",
    ) -> dict:
        """Builds a dictionary for updating or creating an entity in Weclapp.
        Args:
            update_type (Literal[full, used]): Type of update. Defaults to
                "used".
                - "full": Update dictionary will include all attributes of the
                    instance. This type of update should be used with caution as it
                    may lead to optimistic lock errors and could potentially overwrite
                    some values to empty if they are not set in the instance.
                - "used": Update dictionary will include only attributes that have been used.
        Returns:
            dict: Dictionary representing the entity to be updated or created.
        """
        data_to_send = {}
        if self.updated or update_type == "full":
            data_to_send["attributeDefinitionId"] = self.attributeDefinitionId
            if self.valueName is not None:
                data_to_send[self.valueName] = self.value
        return data_to_send



"""This code was dynamically created using WeclappClassCreator from pyweclapp"""

from typing import Union, List, Set
from .blueprints import Blueprint, WeclappMetaData


class InvoiceAddress(Blueprint):
    city: Union[str, None] = None
    company: Union[str, None] = None
    company2: Union[str, None] = None
    countryCode: Union[str, None] = None
    firstName: Union[str, None] = None
    globalLocationNumber: Union[str, None] = None
    lastName: Union[str, None] = None
    middleName: Union[str, None] = None
    phoneNumber: Union[str, None] = None
    postOfficeBoxCity: Union[str, None] = None
    postOfficeBoxNumber: Union[str, None] = None
    postOfficeBoxZipCode: Union[str, None] = None
    salutation: Union[str, None] = None
    state: Union[str, None] = None
    street1: Union[str, None] = None
    street2: Union[str, None] = None
    titleId: Union[str, None] = None
    zipcode: Union[str, None] = None


class RecipientAddress(Blueprint):
    city: Union[str, None] = None
    company: Union[str, None] = None
    company2: Union[str, None] = None
    countryCode: Union[str, None] = None
    firstName: Union[str, None] = None
    globalLocationNumber: Union[str, None] = None
    lastName: Union[str, None] = None
    middleName: Union[str, None] = None
    phoneNumber: Union[str, None] = None
    postOfficeBoxCity: Union[str, None] = None
    postOfficeBoxNumber: Union[str, None] = None
    postOfficeBoxZipCode: Union[str, None] = None
    salutation: Union[str, None] = None
    state: Union[str, None] = None
    street1: Union[str, None] = None
    street2: Union[str, None] = None
    titleId: Union[str, None] = None
    zipcode: Union[str, None] = None


class SalesOrders(Blueprint):
    id: Union[str, None] = None


class StatusHistory(Blueprint):
    status: Union[str, None] = None
    statusDate: Union[int, None] = None
    userId: Union[str, None] = None


class PurchaseOrders(Blueprint):
    id: Union[str, None] = None


class RecordEmailAddresses(Blueprint):
    bccAddresses: Union[str, None] = None
    ccAddresses: Union[str, None] = None
    toAddresses: Union[str, None] = None


class SalesInvoiceEmailAddresses(Blueprint):
    bccAddresses: Union[str, None] = None
    ccAddresses: Union[str, None] = None
    toAddresses: Union[str, None] = None


class Picks(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    batchNumber: Union[str, None] = None
    confirmedByUserId: Union[str, None] = None
    confirmedDate: Union[int, None] = None
    internalTransportReferenceId: Union[str, None] = None
    quantity: Union[str, None] = None
    serialNumbers: list = []
    storagePlaceId: Union[str, None] = None
    bookedDate: Union[int, None] = None
    orderItemId: Union[str, None] = None
    sourceInternalTransportReferenceId: Union[str, None] = None
    sourceStoragePlaceId: Union[str, None] = None
    transportationOrderId: Union[str, None] = None


class ShipmentItems(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    customAttributes: List[WeclappMetaData] = []
    articleId: Union[str, None] = None
    note: Union[str, None] = None
    positionNumber: Union[int, None] = None
    quantity: Union[str, None] = None
    description: Union[str, None] = None
    descriptionFixed: Union[bool, None] = None
    itemType: Union[str, None] = None
    manualQuantity: Union[bool, None] = None
    parentItemId: Union[str, None] = None
    title: Union[str, None] = None
    unitId: Union[str, None] = None
    addPageBreakBefore: Union[bool, None] = None
    groupName: Union[str, None] = None
    picks: List[Picks] = []
    purchaseOrderItemId: Union[str, None] = None
    returnAssessmentId: Union[str, None] = None
    returnDescription: Union[str, None] = None
    returnErrorId: Union[str, None] = None
    returnReasonId: Union[str, None] = None
    returnRectificationId: Union[str, None] = None
    salesOrderItemId: Union[str, None] = None

    excluded_keys: Set[str] = {
        "version"
    }


class ShippedFromAddress(Blueprint):
    city: Union[str, None] = None
    company: Union[str, None] = None
    company2: Union[str, None] = None
    countryCode: Union[str, None] = None
    firstName: Union[str, None] = None
    globalLocationNumber: Union[str, None] = None
    lastName: Union[str, None] = None
    middleName: Union[str, None] = None
    phoneNumber: Union[str, None] = None
    postOfficeBoxCity: Union[str, None] = None
    postOfficeBoxNumber: Union[str, None] = None
    postOfficeBoxZipCode: Union[str, None] = None
    salutation: Union[str, None] = None
    state: Union[str, None] = None
    street1: Union[str, None] = None
    street2: Union[str, None] = None
    titleId: Union[str, None] = None
    zipcode: Union[str, None] = None


class Shipment(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    customAttributes: List[WeclappMetaData] = []
    commercialLanguage: Union[str, None] = None
    creatorId: Union[str, None] = None
    description: Union[str, None] = None
    disableRecordEmailingRule: Union[bool, None] = None
    recordComment: Union[str, None] = None
    recordFreeText: Union[str, None] = None
    recordOpening: Union[str, None] = None
    sentToRecipient: Union[bool, None] = None
    tags: list = []
    invoiceAddress: InvoiceAddress = InvoiceAddress()
    recipientAddress: RecipientAddress = RecipientAddress()
    salesOrders: List[SalesOrders] = []
    status: Union[str, None] = None
    statusHistory: List[StatusHistory] = []
    additionalDeliveryInformation: Union[str, None] = None
    consolidationStoragePlaceId: Union[str, None] = None
    customerPurchaseOrderNumber: Union[str, None] = None
    declaredValueAmount: Union[str, None] = None
    declaredValueAmountCurrencyId: Union[str, None] = None
    deliveryDate: Union[int, None] = None
    destinationStoragePlaceId: Union[str, None] = None
    destinationWarehouseId: Union[str, None] = None
    dhlReceiverId: Union[str, None] = None
    invoiceRecipientId: Union[str, None] = None
    mainSalesOrderId: Union[str, None] = None
    packageHeight: Union[int, None] = None
    packageLength: Union[int, None] = None
    packageReferenceNumber: Union[str, None] = None
    packageReturnTrackingNumber: Union[str, None] = None
    packageReturnTrackingUrl: Union[str, None] = None
    packageTrackingNumber: Union[str, None] = None
    packageTrackingUrl: Union[str, None] = None
    packageWeight: Union[str, None] = None
    packageWidth: Union[int, None] = None
    pickingInstructions: Union[str, None] = None
    picksComplete: Union[bool, None] = None
    purchaseOrders: List[PurchaseOrders] = []
    recipientCustomerNumber: Union[str, None] = None
    recipientPartyId: Union[str, None] = None
    recipientSupplierNumber: Union[str, None] = None
    recordEmailAddresses: RecordEmailAddresses = None
    responsibleUserId: Union[str, None] = None
    salesInvoiceEmailAddresses: SalesInvoiceEmailAddresses = SalesInvoiceEmailAddresses()
    shipmentItems: List[ShipmentItems] = []
    shipmentMethodId: Union[str, None] = None
    shipmentNumber: Union[str, None] = None
    shipmentType: Union[str, None] = None
    shippedFromAddress: ShippedFromAddress = ShippedFromAddress()
    shippingCarrierId: Union[str, None] = None
    shippingDate: Union[int, None] = None
    shippingLabelsCount: Union[int, None] = None
    shippingReturnCarrierId: Union[str, None] = None
    warehouseId: Union[str, None] = None

    excluded_keys: Set[str] = {
        "picksComplete"
    }



"""This code was dynamically created using WeclappClassCreator from pyweclapp"""

from typing import Union
from .blueprints import Blueprint


class Currency(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    currencySymbol: Union[str, None] = None
    name: Union[str, None] = None



"""Module to create Weclapp Class Blueprints."""

import os
import re
from typing import Any
from ...weclapp import Weclapp
from . import config


class WeclappClassCreator:
    """Class to create Weclapp Class Blueprints.
    Args:
        entity_name (str): Name of the entity to create a class for.
        target_directory (str): Directory where the class file should be created.
        entity_dict (dict, optional): Dictionary containing the entity data. If
            not provided, it will be fetched from the Weclapp API.

    After initialization call create_python_file() to generate the class file.
    """

    def __init__(
        self, entity_name: str, target_directory: str, entity_dict: dict = None
    ):
        if entity_dict is not None:
            self.entity = entity_dict
        else:
            self.entity = self.get_example_entity(entity_name)
        self.entity_name = entity_name
        self.target_directory = target_directory
        self.class_templates = []

    def get_example_entity(self, entity_name: str) -> dict:
        """Fetches an example entity from the Weclapp API to create a class
        blueprint."""
        weclapp = Weclapp()
        entities = weclapp.get(entity_name, query=config.CREATION_QUERY, as_type=list)
        if entities and isinstance(entities, list) and len(entities) > 0:
            return entities[0]

        raise ValueError(
            f"Could not find entity {entity_name} in Weclapp API. "
            "Please check the entity name or provide a valid entity dictionary."
            f"Response: {entities}"
        )

    def create_python_file(self) -> None:
        """Main function to generate the python file."""
        if not os.path.exists(self.target_directory):
            os.makedirs(self.target_directory)

        self.create_class_templates(self.entity_name, self.entity)
        file_content = config.STATIC_IMPORTS_MODEL_FILES

        file_content += "\n\n".join(self.class_templates)

        splitted_entity_name = re.split(r"(?=[A-Z])", self.entity_name)
        file_name = f"{'_'.join(splitted_entity_name).lower()}_model"
        with open(
            f"{self.target_directory}/{file_name}.py", "w+", encoding="utf-8"
        ) as file:
            file.write(file_content)

        self.update_init_file(file_name)

    def create_class_templates(self, entity_name: str, entity: dict) -> str:
        """Generates the python class code string for a given entity"""
        class_name = self.capitalize(entity_name)
        file_content = f"class {class_name}(Blueprint):\n"

        for key, value in entity.items():
            if key == "customAttributes":
                file_content += f"    {key}: List[WeclappMetaData] = []\n"

            elif isinstance(value, list):
                if len(value) > 0 and isinstance(value[0], dict):
                    child_name = self.create_class_templates(key, value[0])
                    file_content += f"    {key}: List[{child_name}] = []\n"
                else:
                    file_content += f"    {key}: list = []\n"

            elif isinstance(value, dict):
                child_name = self.create_class_templates(key, value)
                file_content += f"    {key}: {child_name} = {child_name}()\n"

            else:
                file_content += self.generate_type_hint(value, key)

        self.class_templates.append(file_content)
        return class_name

    def capitalize(self, string: str) -> str:
        """Capitalizes the first letter of a given string"""
        return string[:1].upper() + string[1:]

    def generate_type_hint(self, value: Any, key: str) -> str:
        """Generate type hint for a given value."""
        attribute_type = type(value)
        if attribute_type.__name__ == "NoneType":
            estimated_type = "int" if "date" in str(key).lower() else "str"
            return f"    {key}: Optional[{estimated_type}] = None  # Type estimated\n"
        else:
            relevant_type = attribute_type.__name__
            return f"    {key}: Union[{relevant_type}, None] = None\n"

    def update_init_file(self, file_name: str):
        """Imports the generated entity to the __init__.py file of the target
        directory."""
        init_path = os.path.join(self.target_directory, config.INIT_FILE_NAME)
        if not os.path.exists(init_path):
            with open(init_path, "w+", encoding="utf-8") as file:
                file.write("")

        current_imports = []
        with open(init_path, "r+", encoding="utf-8") as file:
            for line in file.readlines():
                if line.startswith("from ."):
                    current_imports.append(line.strip("\n"))

            new_import = f"from .{file_name} import {self.capitalize(self.entity_name)}"
            if not any(new_import in import_line for import_line in current_imports):
                current_imports.append(new_import)

        with open(init_path, "w+", encoding="utf-8") as file:
            current_imports.insert(0, config.INIT_FILE_DOC_STRING)
            content = "\n".join(current_imports) + "\n"
            file.write(content)



"""This code was dynamically created using WeclappClassCreator from pyweclapp"""

from typing import Union, List
from .blueprints import Blueprint, WeclappMetaData


class Addresses(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    city: Union[str, None] = None
    company: Union[str, None] = None
    company2: Union[str, None] = None
    countryCode: Union[str, None] = None
    deliveryAddress: Union[bool, None] = None
    firstName: Union[str, None] = None
    globalLocationNumber: Union[str, None] = None
    invoiceAddress: Union[bool, None] = None
    lastName: Union[str, None] = None
    phoneNumber: Union[str, None] = None
    postOfficeBoxCity: Union[str, None] = None
    postOfficeBoxNumber: Union[str, None] = None
    postOfficeBoxZipCode: Union[str, None] = None
    primaryAddress: Union[bool, None] = None
    salutation: Union[str, None] = None
    state: Union[str, None] = None
    street1: Union[str, None] = None
    street2: Union[str, None] = None
    titleId: Union[str, None] = None
    zipcode: Union[str, None] = None


class OnlineAccounts(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    accountName: Union[str, None] = None
    accountType: Union[str, None] = None
    url: Union[str, None] = None


class BankAccounts(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    accountHolder: Union[str, None] = None
    accountNumber: Union[str, None] = None
    bankCode: Union[str, None] = None
    creditInstitute: Union[str, None] = None
    partyId: Union[str, None] = None
    primary: Union[bool, None] = None


class CommissionSalesPartners(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    commissionFix: Union[str, None] = None
    commissionPercentage: Union[str, None] = None
    commissionType: Union[str, None] = None
    salesPartnerSupplierId: Union[str, None] = None


class Contacts(Blueprint):
    id: Union[str, None] = None


class CustomerSalesStageHistory(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    salesStageId: Union[str, None] = None
    userId: Union[str, None] = None


class PartyEmailAddresses(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    bccAddresses: Union[str, None] = None
    ccAddresses: Union[str, None] = None
    toAddresses: Union[str, None] = None


class Invoices(Blueprint):
    id: Union[str, None] = None


class PartyHabitualExporterLettersOfIntent(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    automaticallySuggestInInvoice: Union[bool, None] = None
    date: Union[int, None] = None
    fromSupplier: Union[bool, None] = None
    invoices: List[Invoices] = []
    numberDeclarer: Union[str, None] = None
    numberSupplier: Union[str, None] = None
    totalAmount: Union[str, None] = None
    type: Union[str, None] = None


class Topics(Blueprint):
    id: Union[str, None] = None


class Party(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    customAttributes: List[WeclappMetaData] = []
    addresses: List[Addresses] = []
    birthDate: Union[int, None] = None
    company: Union[str, None] = None
    company2: Union[str, None] = None
    deliveryAddressId: Union[str, None] = None
    email: Union[str, None] = None
    fax: Union[str, None] = None
    firstName: Union[str, None] = None
    invoiceAddressId: Union[str, None] = None
    lastName: Union[str, None] = None
    middleName: Union[str, None] = None
    mobilePhone1: Union[str, None] = None
    onlineAccounts: List[OnlineAccounts] = []
    partyType: Union[str, None] = None
    personCompany: Union[str, None] = None
    personDepartmentId: Union[str, None] = None
    personRoleId: Union[str, None] = None
    phone: Union[str, None] = None
    primaryAddressId: Union[str, None] = None
    salutation: Union[str, None] = None
    tags: list = []
    titleId: Union[str, None] = None
    website: Union[str, None] = None
    bankAccounts: List[BankAccounts] = []
    commercialLanguageId: Union[str, None] = None
    commissionBlock: Union[bool, None] = None
    commissionSalesPartners: List[CommissionSalesPartners] = []
    companySizeId: Union[str, None] = None
    competitor: Union[bool, None] = None
    contacts: List[Contacts] = []
    convertedOnDate: Union[int, None] = None
    currencyId: Union[str, None] = None
    customer: Union[bool, None] = None
    customerAllowDropshippingOrderCreation: Union[bool, None] = None
    customerAmountInsured: Union[str, None] = None
    customerAnnualRevenue: Union[str, None] = None
    customerBlockNotice: Union[str, None] = None
    customerBlocked: Union[bool, None] = None
    customerBusinessType: Union[str, None] = None
    customerCategoryId: Union[str, None] = None
    customerCreditLimit: Union[str, None] = None
    customerCurrentSalesStageId: Union[str, None] = None
    customerDebtorAccountId: Union[str, None] = None
    customerDebtorAccountingCodeId: Union[str, None] = None
    customerDefaultHeaderDiscount: Union[str, None] = None
    customerDefaultHeaderSurcharge: Union[str, None] = None
    customerDefaultShippingCarrierId: Union[str, None] = None
    customerDeliveryBlock: Union[bool, None] = None
    customerInsolvent: Union[bool, None] = None
    customerInsured: Union[bool, None] = None
    customerInternalNote: Union[str, None] = None
    customerLossDescription: Union[str, None] = None
    customerLossReasonId: Union[str, None] = None
    customerNonStandardTaxId: Union[str, None] = None
    customerNumber: Union[str, None] = None
    customerNumberOld: Union[str, None] = None
    customerPaymentMethodId: Union[str, None] = None
    customerSalesChannel: Union[str, None] = None
    customerSalesOrderPaymentType: Union[str, None] = None
    customerSalesProbability: Union[int, None] = None
    customerSalesStageHistory: List[CustomerSalesStageHistory] = []
    customerSatisfaction: Union[str, None] = None
    customerShipmentMethodId: Union[str, None] = None
    customerSupplierNumber: Union[str, None] = None
    customerTermOfPaymentId: Union[str, None] = None
    customerUseCustomsTariffNumber: Union[bool, None] = None
    deliveryEmailAddressesId: Union[str, None] = None
    description: Union[str, None] = None
    dunningAddressId: Union[str, None] = None
    dunningEmailAddressesId: Union[str, None] = None
    enableDropshippingInNewSupplySources: Union[bool, None] = None
    eoriNumber: Union[str, None] = None
    factoring: Union[bool, None] = None
    fixPhone2: Union[str, None] = None
    fixedResponsibleUser: Union[bool, None] = None
    formerSalesPartner: Union[bool, None] = None
    habitualExporter: Union[bool, None] = None
    imageId: Union[str, None] = None
    invoiceBlock: Union[bool, None] = None
    invoiceRecipientId: Union[str, None] = None
    leadRatingId: Union[str, None] = None
    leadSourceId: Union[str, None] = None
    leadStatus: Union[str, None] = None
    legalFormId: Union[str, None] = None
    mobilePhone2: Union[str, None] = None
    optInEmail: Union[bool, None] = None
    optInLetter: Union[bool, None] = None
    optInPhone: Union[bool, None] = None
    optInSms: Union[bool, None] = None
    parentPartyId: Union[str, None] = None
    partyEmailAddresses: List[PartyEmailAddresses] = []
    partyHabitualExporterLettersOfIntent: List[PartyHabitualExporterLettersOfIntent] = []
    phoneHome: Union[str, None] = None
    primaryContactId: Union[str, None] = None
    publicPageExpirationDate: Union[int, None] = None
    publicPageUuid: Union[str, None] = None
    purchaseEmailAddressesId: Union[str, None] = None
    purchaseViaPlafond: Union[bool, None] = None
    quotationEmailAddressesId: Union[str, None] = None
    ratingId: Union[str, None] = None
    referenceNumber: Union[str, None] = None
    regionId: Union[str, None] = None
    responsibleUserId: Union[str, None] = None
    salesInvoiceEmailAddressesId: Union[str, None] = None
    salesOrderEmailAddressesId: Union[str, None] = None
    salesPartner: Union[bool, None] = None
    salesPartnerDefaultCommissionFix: Union[str, None] = None
    salesPartnerDefaultCommissionPercentage: Union[str, None] = None
    salesPartnerDefaultCommissionType: Union[str, None] = None
    sectorId: Union[str, None] = None
    supplier: Union[bool, None] = None
    supplierActive: Union[bool, None] = None
    supplierCreditorAccountId: Union[str, None] = None
    supplierCreditorAccountingCodeId: Union[str, None] = None
    supplierCustomerNumberAtSupplier: Union[str, None] = None
    supplierDefaultShippingCarrierId: Union[str, None] = None
    supplierInternalNote: Union[str, None] = None
    supplierMergeItemsForOcrInvoiceUpload: Union[bool, None] = None
    supplierMinimumPurchaseOrderAmount: Union[str, None] = None
    supplierNonStandardTaxId: Union[str, None] = None
    supplierNumber: Union[str, None] = None
    supplierNumberOld: Union[str, None] = None
    supplierOrderBlock: Union[bool, None] = None
    supplierPaymentMethodId: Union[str, None] = None
    supplierShipmentMethodId: Union[str, None] = None
    supplierTermOfPaymentId: Union[str, None] = None
    taxId: Union[str, None] = None
    topics: List[Topics] = []
    vatIdentificationNumber: Union[str, None] = None
    xRechnungLeitwegId: Union[str, None] = None



"""This code was dynamically created using WeclappClassCreator from pyweclapp"""

from typing import Union, List, Set
from .blueprints import Blueprint, WeclappMetaData


class ArticleAlternativeQuantities(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    minimumOrderQuantity: Union[str, None] = None
    minimumStockQuantity: Union[str, None] = None
    targetStockQuantity: Union[str, None] = None
    warehouseId: Union[str, None] = None


class ArticleCalculationPrices(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    articleCalculationPriceType: Union[str, None] = None
    endDate: Union[int, None] = None
    price: Union[str, None] = None
    salesChannel: Union[str, None] = None
    startDate: Union[int, None] = None


class ArticleImages(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    fileName: Union[str, None] = None
    mainImage: Union[bool, None] = None


class ReductionAdditions(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    name: Union[str, None] = None
    type: Union[str, None] = None
    value: Union[str, None] = None


class ArticlePrices(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    currencyId: Union[str, None] = None
    customerId: Union[str, None] = None
    description: Union[str, None] = None
    endDate: Union[int, None] = None
    lastModifiedByUserId: Union[str, None] = None
    price: Union[str, None] = None
    priceScaleType: Union[str, None] = None
    priceScaleValue: Union[str, None] = None
    reductionAdditions: List[ReductionAdditions] = []
    startDate: Union[int, None] = None
    salesChannel: Union[str, None] = None


class CustomerArticleNumbers(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    customerArticleNumber: Union[str, None] = None
    customerId: Union[str, None] = None


class DefaultStoragePlaces(Blueprint):
    id: Union[str, None] = None


class ProductionBillOfMaterialItems(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    articleId: Union[str, None] = None
    positionNumber: Union[int, None] = None
    quantity: Union[str, None] = None


class QuantityConversions(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    conversionQuantity: Union[str, None] = None
    createdUserId: Union[str, None] = None
    lastEditedUserId: Union[str, None] = None
    oppositeDirection: Union[bool, None] = None
    unitId: Union[str, None] = None


class SalesBillOfMaterialItems(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    articleId: Union[str, None] = None
    positionNumber: Union[int, None] = None
    quantity: Union[str, None] = None


class SupplySources(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    articleSupplySourceId: Union[str, None] = None
    positionNumber: Union[int, None] = None


class Article(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    customAttributes: List[WeclappMetaData] = []
    articleNumber: Union[str, None] = None
    description: Union[str, None] = None
    ean: Union[str, None] = None
    fixedPurchaseQuantity: Union[str, None] = None
    internalNote: Union[str, None] = None
    manufacturerPartNumber: Union[str, None] = None
    matchCode: Union[str, None] = None
    minimumPurchaseQuantity: Union[str, None] = None
    name: Union[str, None] = None
    shortDescription1: Union[str, None] = None
    shortDescription2: Union[str, None] = None
    taxRateType: Union[str, None] = None
    unitId: Union[str, None] = None
    accountId: Union[str, None] = None
    accountingCodeId: Union[str, None] = None
    active: Union[bool, None] = None
    applyCashDiscount: Union[bool, None] = None
    articleAlternativeQuantities: List[ArticleAlternativeQuantities] = []
    articleCalculationPrices: List[ArticleCalculationPrices] = []
    articleCategoryId: Union[str, None] = None
    articleGrossWeight: Union[str, None] = None
    articleHeight: Union[str, None] = None
    articleImages: List[ArticleImages] = []
    articleLength: Union[str, None] = None
    articleNetWeight: Union[str, None] = None
    articlePrices: List[ArticlePrices] = []
    articleType: Union[str, None] = None
    articleWidth: Union[str, None] = None
    availableForSalesChannels: list = []
    availableInSale: Union[bool, None] = None
    averageDeliveryTime: Union[int, None] = None
    barcode: Union[str, None] = None
    batchNumberRequired: Union[bool, None] = None
    billOfMaterialPartDeliveryPossible: Union[bool, None] = None
    catalogCode: Union[str, None] = None
    commissionRate: Union[str, None] = None
    contractBillingCycle: Union[str, None] = None
    contractBillingMode: Union[str, None] = None
    countryOfOriginCode: Union[str, None] = None
    customerArticleNumbers: List[CustomerArticleNumbers] = []
    customsDescription: Union[str, None] = None
    customsTariffNumberId: Union[str, None] = None
    defaultLoadingEquipmentIdentifierId: Union[str, None] = None
    defaultPriceCalculationType: Union[str, None] = None
    defaultStoragePlaces: List[DefaultStoragePlaces] = []
    defineIndividualTaskTemplates: Union[bool, None] = None
    expenseAccountId: Union[str, None] = None
    expirationDays: Union[int, None] = None
    invoicingType: Union[str, None] = None
    launchDate: Union[int, None] = None
    loadingEquipmentArticleId: Union[str, None] = None
    longText: Union[str, None] = None
    lowLevelCode: Union[int, None] = None
    manufacturerId: Union[str, None] = None
    marginCalculationPriceType: Union[str, None] = None
    minimumStockQuantity: Union[str, None] = None
    packagingQuantity: Union[int, None] = None
    packagingUnitBaseArticleId: Union[str, None] = None
    packagingUnitParentArticleId: Union[str, None] = None
    plannedWorkingTimePerUnit: Union[int, None] = None
    primarySupplySourceId: Union[str, None] = None
    procurementLeadDays: Union[int, None] = None
    producerType: Union[str, None] = None
    productionArticle: Union[bool, None] = None
    productionBillOfMaterialItems: List[ProductionBillOfMaterialItems] = []
    productionConfigurationRule: Union[str, None] = None
    purchaseCostCenterId: Union[str, None] = None
    quantityConversions: List[QuantityConversions] = []
    ratingId: Union[str, None] = None
    recordItemGroupName: Union[str, None] = None
    safetyStockDays: Union[int, None] = None
    salesBillOfMaterialItems: List[SalesBillOfMaterialItems] = []
    salesCostCenterId: Union[str, None] = None
    sellByDate: Union[int, None] = None
    sellFromDate: Union[int, None] = None
    serialNumberRequired: Union[bool, None] = None
    serviceArticleForServiceQuotaBookingId: Union[str, None] = None
    serviceQuotaQuantity: Union[str, None] = None
    showOnDeliveryNote: Union[bool, None] = None
    statusId: Union[str, None] = None
    supplySources: List[SupplySources] = []
    supportUntilDate: Union[int, None] = None
    systemCode: Union[str, None] = None
    tags: list = []
    targetStockQuantity: Union[str, None] = None
    useAvailableForSalesChannels: Union[bool, None] = None
    useSalesBillOfMaterialItemPrices: Union[bool, None] = None
    useSalesBillOfMaterialItemPricesForPurchase: Union[bool, None] = None
    useSalesBillOfMaterialSubitemCosts: Union[bool, None] = None

    excluded_keys: Set[str] = {
        "lowLevelCode"
    }



from .weclapp import Weclapp, WeclappError
from . import weclapp_classes
from .weclapp_document import DocManager, Document
from . import time_functions
from .custom_attributes import CustomAttributeClass, CustomAttributeGenerator



"""Model for extraInfoForApp endpoint that provides stock and sales information
for articles."""

from pydantic import BaseModel
from ..weclapp import Weclapp


class ExtraInfoForApp(BaseModel):
    """Model for article/id/{articleId}/extraInfoForApp endpoint."""
    confirmedOrderedQuantity: str = None
    consignmentStockQuantity: str = None
    currentYearRevenue: str = None
    currentYearSalesOrderVolume: str = None
    inventoryValuationAtAvgArticlePrice: str = None
    lastYearRevenue: str = None
    lastYearSalesOrderVolume: str = None
    openQuantityInShipments: str = None
    openShipmentQuantity: str = None
    orderedQuantity: str = None
    plannedSalesQuantity: str = None
    purchasePrice: str = None
    purchasePriceCurrencyId: str = None
    reservedStockQuantity: str = None
    salesPrice: str = None
    salesPriceCurrencyId: str = None
    stockQuantity: str = None
    stockQuantityWithoutOrder: str = None
    stockQuantityWithoutOrderToDate: str = None
    stockValuationPrice: str = None
    supplierStockQuantity: str = None
    unconfirmedOrderedQuantity: str = None
    unreservedStockQuantity: str = None

    def __init__(self, **kwargs):
        BaseModel.__init__(self, **kwargs)

    @classmethod
    def from_weclapp(cls, article_id: str):
        """Fetch extra info for an article from Weclapp."""
        response = Weclapp(api_version="v1").get(f"article/id/{article_id}/extraInfoForApp")
        return cls(**response)



"""Module for weclapp classes package imports. Add more classes if necessary."""

from .blueprints.custom_attributes_model import WeclappMetaData
from .sales_order_model import SalesOrder, OrderItems
from .incoming_goods_model import IncomingGoods, IncomingGoodsItems
from .purchase_order_model import PurchaseOrder, PurchaseOrderItems
from .sales_invoice_model import SalesInvoice, SalesInvoiceItems
from .shipment_model import Shipment, ShipmentItems
from .quotation_model import Quotation
from .party_model import Party
from .extra_info_for_app_model import ExtraInfoForApp
from .article_model import Article, ArticlePrices
from .currency_model import Currency



"""This code was dynamically created using WeclappClassCreator from pyweclapp"""

from typing import Union, List, Set
from .blueprints import Blueprint, WeclappMetaData


class CommissionSalesPartners(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    commissionFix: Union[str, None] = None
    commissionPercentage: Union[str, None] = None
    commissionType: Union[str, None] = None
    salesPartnerSupplierId: Union[str, None] = None


class DeliveryAddress(Blueprint):
    city: Union[str, None] = None
    company: Union[str, None] = None
    company2: Union[str, None] = None
    countryCode: Union[str, None] = None
    firstName: Union[str, None] = None
    globalLocationNumber: Union[str, None] = None
    lastName: Union[str, None] = None
    middleName: Union[str, None] = None
    phoneNumber: Union[str, None] = None
    postOfficeBoxCity: Union[str, None] = None
    postOfficeBoxNumber: Union[str, None] = None
    postOfficeBoxZipCode: Union[str, None] = None
    salutation: Union[str, None] = None
    state: Union[str, None] = None
    street1: Union[str, None] = None
    street2: Union[str, None] = None
    titleId: Union[str, None] = None
    zipcode: Union[str, None] = None


class DeliveryEmailAddresses(Blueprint):
    bccAddresses: Union[str, None] = None
    ccAddresses: Union[str, None] = None
    toAddresses: Union[str, None] = None


class InvoiceAddress(Blueprint):
    city: Union[str, None] = None
    company: Union[str, None] = None
    company2: Union[str, None] = None
    countryCode: Union[str, None] = None
    firstName: Union[str, None] = None
    globalLocationNumber: Union[str, None] = None
    lastName: Union[str, None] = None
    middleName: Union[str, None] = None
    phoneNumber: Union[str, None] = None
    postOfficeBoxCity: Union[str, None] = None
    postOfficeBoxNumber: Union[str, None] = None
    postOfficeBoxZipCode: Union[str, None] = None
    salutation: Union[str, None] = None
    state: Union[str, None] = None
    street1: Union[str, None] = None
    street2: Union[str, None] = None
    titleId: Union[str, None] = None
    zipcode: Union[str, None] = None


class RecordAddress(Blueprint):
    city: Union[str, None] = None
    company: Union[str, None] = None
    company2: Union[str, None] = None
    countryCode: Union[str, None] = None
    firstName: Union[str, None] = None
    globalLocationNumber: Union[str, None] = None
    lastName: Union[str, None] = None
    middleName: Union[str, None] = None
    phoneNumber: Union[str, None] = None
    postOfficeBoxCity: Union[str, None] = None
    postOfficeBoxNumber: Union[str, None] = None
    postOfficeBoxZipCode: Union[str, None] = None
    salutation: Union[str, None] = None
    state: Union[str, None] = None
    street1: Union[str, None] = None
    street2: Union[str, None] = None
    titleId: Union[str, None] = None
    zipcode: Union[str, None] = None


class SalesInvoiceEmailAddresses(Blueprint):
    bccAddresses: Union[str, None] = None
    ccAddresses: Union[str, None] = None
    toAddresses: Union[str, None] = None


class EcommerceOrder(Blueprint):
    ecommerceId: Union[str, None] = None
    externalConnectionId: Union[str, None] = None


class ReductionAdditionItems(Blueprint):
    position: Union[int, None] = None
    source: Union[str, None] = None
    specialPriceReduction: Union[bool, None] = None
    title: Union[str, None] = None
    type: Union[str, None] = None
    value: Union[str, None] = None


class Picks(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    batchNumber: Union[str, None] = None
    confirmedByUserId: Union[str, None] = None
    confirmedDate: Union[int, None] = None
    internalTransportReferenceId: Union[str, None] = None
    quantity: Union[str, None] = None
    serialNumbers: list = []
    storagePlaceId: Union[str, None] = None
    bookedDate: Union[int, None] = None
    orderItemId: Union[str, None] = None
    sourceInternalTransportReferenceId: Union[str, None] = None
    sourceStoragePlaceId: Union[str, None] = None
    transportationOrderId: Union[str, None] = None


class Tasks(Blueprint):
    id: Union[str, None] = None


class OrderItems(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    customAttributes: List[WeclappMetaData] = []
    articleId: Union[str, None] = None
    note: Union[str, None] = None
    positionNumber: Union[int, None] = None
    quantity: Union[str, None] = None
    description: Union[str, None] = None
    descriptionFixed: Union[bool, None] = None
    itemType: Union[str, None] = None
    manualQuantity: Union[bool, None] = None
    parentItemId: Union[str, None] = None
    title: Union[str, None] = None
    unitId: Union[str, None] = None
    discountPercentage: Union[str, None] = None
    grossAmount: Union[str, None] = None
    grossAmountInCompanyCurrency: Union[str, None] = None
    manualUnitPrice: Union[bool, None] = None
    netAmount: Union[str, None] = None
    netAmountForStatistics: Union[str, None] = None
    netAmountForStatisticsInCompanyCurrency: Union[str, None] = None
    netAmountInCompanyCurrency: Union[str, None] = None
    reductionAdditionItems: List[ReductionAdditionItems] = []
    taxId: Union[str, None] = None
    unitPrice: Union[str, None] = None
    unitPriceInCompanyCurrency: Union[str, None] = None
    addPageBreakBefore: Union[bool, None] = None
    groupName: Union[str, None] = None
    commissionSalesPartners: List[CommissionSalesPartners] = []
    manualUnitCost: Union[bool, None] = None
    recommendedRetailPrice: Union[str, None] = None
    servicePeriodFrom: Union[int, None] = None
    servicePeriodTo: Union[int, None] = None
    unitCost: Union[str, None] = None
    unitCostInCompanyCurrency: Union[str, None] = None
    invoicingType: Union[str, None] = None
    manualPlannedWorkingTimePerUnit: Union[bool, None] = None
    plannedDeliveryDate: Union[int, None] = None
    plannedShippingDate: Union[int, None] = None
    plannedWorkingTimePerUnit: Union[int, None] = None
    contractChargeId: Union[str, None] = None
    ecommerceOrderItemIds: list = []
    invoicedQuantity: Union[str, None] = None
    picks: List[Picks] = []
    returnedQuantity: Union[str, None] = None
    serviceQuotaId: Union[str, None] = None
    shipped: Union[bool, None] = None
    shippedQuantity: Union[str, None] = None
    tasks: List[Tasks] = []

    excluded_keys: Set[str] = {
        "version",
        "grossAmount",
        "grossAmountInCompanyCurrency",
        "netAmount",
        "netAmountForStatistics",
        "netAmountForStatisticsInCompanyCurrency",
        "netAmountInCompanyCurrency",
        "unitPriceInCompanyCurrency",
        "unitCostInCompanyCurrency",
        "invoicedQuantity",
        "shippedQuantity",
        "returnedQuantity",
        "shipped",
    }


class SalesInvoices(Blueprint):
    id: Union[str, None] = None


class Payments(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    amount: Union[str, None] = None
    condition: Union[str, None] = None
    conditionMet: Union[bool, None] = None
    dueDate: Union[int, None] = None
    positionNumber: Union[int, None] = None
    salesInvoiceId: Union[str, None] = None
    salesInvoices: List[SalesInvoices] = []


class ProjectMembers(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    hourlyCost: Union[str, None] = None
    teamRole: Union[str, None] = None
    userId: Union[str, None] = None


class RecordEmailAddresses(Blueprint):
    bccAddresses: Union[str, None] = None
    ccAddresses: Union[str, None] = None
    toAddresses: Union[str, None] = None


class ShippingCostItems(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    articleId: Union[str, None] = None
    grossAmount: Union[str, None] = None
    grossAmountInCompanyCurrency: Union[str, None] = None
    manualUnitPrice: Union[bool, None] = None
    netAmount: Union[str, None] = None
    netAmountInCompanyCurrency: Union[str, None] = None
    unitPrice: Union[str, None] = None
    unitPriceInCompanyCurrency: Union[str, None] = None
    manualUnitCost: Union[bool, None] = None
    taxId: Union[str, None] = None
    unitCost: Union[str, None] = None
    unitCostInCompanyCurrency: Union[str, None] = None
    ecommerceOrderItemIds: list = []

    excluded_keys: Set[str] = {
        "version",
        "grossAmount",
        "grossAmountInCompanyCurrency",
        "netAmount",
        "netAmountInCompanyCurrency",
        "unitPriceInCompanyCurrency",
        "unitCostInCompanyCurrency",
    }


class StatusHistory(Blueprint):
    creditLimitExceeded: Union[bool, None] = None
    status: Union[str, None] = None
    statusDate: Union[int, None] = None
    userId: Union[str, None] = None


class SalesOrder(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    customAttributes: List[WeclappMetaData] = []
    commercialLanguage: Union[str, None] = None
    creatorId: Union[str, None] = None
    description: Union[str, None] = None
    disableRecordEmailingRule: Union[bool, None] = None
    recordComment: Union[str, None] = None
    recordFreeText: Union[str, None] = None
    recordOpening: Union[str, None] = None
    sentToRecipient: Union[bool, None] = None
    tags: list = []
    currencyConversionDate: Union[int, None] = None
    currencyConversionLocked: Union[bool, None] = None
    currencyConversionRate: Union[str, None] = None
    grossAmount: Union[str, None] = None
    grossAmountInCompanyCurrency: Union[str, None] = None
    headerDiscount: Union[str, None] = None
    headerSurcharge: Union[str, None] = None
    netAmount: Union[str, None] = None
    netAmountInCompanyCurrency: Union[str, None] = None
    nonStandardTaxId: Union[str, None] = None
    paymentMethodId: Union[str, None] = None
    recordCurrencyId: Union[str, None] = None
    termOfPaymentId: Union[str, None] = None
    commission: Union[str, None] = None
    commissionSalesPartners: List[CommissionSalesPartners] = []
    customerId: Union[str, None] = None
    dispatchCountryCode: Union[str, None] = None
    factoring: Union[bool, None] = None
    pricingDate: Union[int, None] = None
    responsibleUserId: Union[str, None] = None
    salesChannel: Union[str, None] = None
    servicePeriodFrom: Union[int, None] = None
    servicePeriodTo: Union[int, None] = None
    shipmentMethodId: Union[str, None] = None
    defaultShippingCarrierId: Union[str, None] = None
    deliveryAddress: DeliveryAddress = DeliveryAddress()
    deliveryEmailAddresses: DeliveryEmailAddresses = DeliveryEmailAddresses()
    invoiceAddress: InvoiceAddress = InvoiceAddress()
    plannedDeliveryDate: Union[int, None] = None
    plannedShippingDate: Union[int, None] = None
    recordAddress: RecordAddress = RecordAddress()
    salesInvoiceEmailAddresses: SalesInvoiceEmailAddresses = SalesInvoiceEmailAddresses()
    advancePaymentAmount: Union[str, None] = None
    advancePaymentStatus: Union[str, None] = None
    applyShippingCostsOnlyOnce: Union[bool, None] = None
    cashAccountId: Union[str, None] = None
    customerHabitualExporterLetterOfIntentId: Union[str, None] = None
    defaultShippingReturnCarrierId: Union[str, None] = None
    ecommerceOrder: EcommerceOrder = EcommerceOrder()
    fulfillmentProviderId: Union[str, None] = None
    invoiceRecipientId: Union[str, None] = None
    invoiced: Union[bool, None] = None
    note: Union[str, None] = None
    onlyServices: Union[bool, None] = None
    orderDate: Union[int, None] = None
    orderItems: List[OrderItems] = []
    orderNumber: Union[str, None] = None
    orderNumberAtCustomer: Union[str, None] = None
    paid: Union[bool, None] = None
    payments: List[Payments] = []
    plannedProjectEndDate: Union[int, None] = None
    plannedProjectStartDate: Union[int, None] = None
    projectGoals: Union[str, None] = None
    projectMembers: List[ProjectMembers] = []
    projectModeActive: Union[bool, None] = None
    quotationId: Union[str, None] = None
    recordAsn: Union[str, None] = None
    recordCommentInheritance: Union[bool, None] = None
    recordEmailAddresses: RecordEmailAddresses = RecordEmailAddresses()
    recordFreeTextInheritance: Union[bool, None] = None
    recordOpeningInheritance: Union[bool, None] = None
    salesOrderPaymentType: Union[str, None] = None
    sepaDirectDebitMandateId: Union[str, None] = None
    servicesFinished: Union[bool, None] = None
    shipped: Union[bool, None] = None
    shippingCostItems: List[ShippingCostItems] = []
    shippingLabelsCount: Union[int, None] = None
    status: Union[str, None] = None
    statusHistory: List[StatusHistory] = []
    template: Union[bool, None] = None
    warehouseId: Union[str, None] = None

    excluded_keys: Set[str] = {
        "paid",
        "unitCostInCompanyCurrency",
        "unitPriceInCompanyCurrency",
        "shippedQuantity",
        "invoicedQuantity",
        "returnedQuantity",
        "shipped",
        "invoiced",
    }



"""Configuration for custom attributes generator."""

STATIC_IDS_JSON_FILE = "special_attributes.json"
JSON_DATA_FILE_NAME = "json_data"

ATTRIBUTE_QUERY = {"active-eq": True}
DEFAULT_GROUP_NAME = "~ no group"

# --------- Generate Python Code ---------
BASE_ATTRIBUTES = ["id", "valueName"]



"""This code was dynamically created using WeclappClassCreator from pyweclapp"""

from typing import Union, List, Set
from .blueprints import Blueprint, WeclappMetaData


class DeliveryAddress(Blueprint):
    city: Union[str, None] = None
    company: Union[str, None] = None
    company2: Union[str, None] = None
    countryCode: Union[str, None] = None
    firstName: Union[str, None] = None
    globalLocationNumber: Union[str, None] = None
    lastName: Union[str, None] = None
    middleName: Union[str, None] = None
    phoneNumber: Union[str, None] = None
    postOfficeBoxCity: Union[str, None] = None
    postOfficeBoxNumber: Union[str, None] = None
    postOfficeBoxZipCode: Union[str, None] = None
    salutation: Union[str, None] = None
    state: Union[str, None] = None
    street1: Union[str, None] = None
    street2: Union[str, None] = None
    titleId: Union[str, None] = None
    zipcode: Union[str, None] = None


class RecordAddress(Blueprint):
    city: Union[str, None] = None
    company: Union[str, None] = None
    company2: Union[str, None] = None
    countryCode: Union[str, None] = None
    firstName: Union[str, None] = None
    globalLocationNumber: Union[str, None] = None
    lastName: Union[str, None] = None
    middleName: Union[str, None] = None
    phoneNumber: Union[str, None] = None
    postOfficeBoxCity: Union[str, None] = None
    postOfficeBoxNumber: Union[str, None] = None
    postOfficeBoxZipCode: Union[str, None] = None
    salutation: Union[str, None] = None
    state: Union[str, None] = None
    street1: Union[str, None] = None
    street2: Union[str, None] = None
    titleId: Union[str, None] = None
    zipcode: Union[str, None] = None


class RecordEmailAddresses(Blueprint):
    bccAddresses: Union[str, None] = None
    ccAddresses: Union[str, None] = None
    toAddresses: Union[str, None] = None


class ReductionAdditionItems(Blueprint):
    position: Union[int, None] = None
    source: Union[str, None] = None
    specialPriceReduction: Union[bool, None] = None
    title: Union[str, None] = None
    type: Union[str, None] = None
    value: Union[str, None] = None


class CommissionSalesPartners(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    commissionFix: Union[str, None] = None
    commissionPercentage: Union[str, None] = None
    commissionType: Union[str, None] = None
    salesPartnerSupplierId: Union[str, None] = None


class CostCenterItems(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    costCenterId: Union[str, None] = None
    distributionPercentage: Union[str, None] = None


class SalesInvoiceItemRelationships(Blueprint):
    performanceRecordItemId: Union[str, None] = None
    quantity: Union[str, None] = None
    salesOrderItemId: Union[str, None] = None
    shipmentItemId: Union[str, None] = None


class SalesInvoiceItems(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    customAttributes: List[WeclappMetaData] = []
    articleId: Union[str, None] = None
    note: Union[str, None] = None
    positionNumber: Union[int, None] = None
    quantity: Union[str, None] = None
    description: Union[str, None] = None
    descriptionFixed: Union[bool, None] = None
    itemType: Union[str, None] = None
    manualQuantity: Union[bool, None] = None
    parentItemId: Union[str, None] = None
    title: Union[str, None] = None
    unitId: Union[str, None] = None
    discountPercentage: Union[str, None] = None
    grossAmount: Union[str, None] = None
    grossAmountInCompanyCurrency: Union[str, None] = None
    manualUnitPrice: Union[bool, None] = None
    netAmount: Union[str, None] = None
    netAmountForStatistics: Union[str, None] = None
    netAmountForStatisticsInCompanyCurrency: Union[str, None] = None
    netAmountInCompanyCurrency: Union[str, None] = None
    reductionAdditionItems: List[ReductionAdditionItems] = []
    taxId: Union[str, None] = None
    unitPrice: Union[str, None] = None
    unitPriceInCompanyCurrency: Union[str, None] = None
    addPageBreakBefore: Union[bool, None] = None
    groupName: Union[str, None] = None
    commissionSalesPartners: List[CommissionSalesPartners] = []
    manualUnitCost: Union[bool, None] = None
    recommendedRetailPrice: Union[str, None] = None
    servicePeriodFrom: Union[int, None] = None
    servicePeriodTo: Union[int, None] = None
    unitCost: Union[str, None] = None
    unitCostInCompanyCurrency: Union[str, None] = None
    accountId: Union[str, None] = None
    contractItemId: Union[str, None] = None
    cost2CostCenterId: Union[str, None] = None
    costCenterItems: List[CostCenterItems] = []
    costTypeId: Union[str, None] = None
    creditedInvoiceItemId: Union[str, None] = None
    deliveryDate: Union[int, None] = None
    salesInvoiceItemRelationships: List[SalesInvoiceItemRelationships] = []
    serialNumbers: list = []
    shippingDate: Union[int, None] = None

    excluded_keys: Set[str] = {
        "version",
        "grossAmount",
        "grossAmountInCompanyCurrency",
        "netAmount",
        "netAmountForStatistics",
        "netAmountForStatisticsInCompanyCurrency",
        "netAmountInCompanyCurrency",
        "unitPriceInCompanyCurrency",
        "unitCostInCompanyCurrency",
        "salesInvoiceItemRelationships",
    }


class SalesOrders(Blueprint):
    id: Union[str, None] = None


class ShippingCostItems(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    articleId: Union[str, None] = None
    grossAmount: Union[str, None] = None
    grossAmountInCompanyCurrency: Union[str, None] = None
    manualUnitPrice: Union[bool, None] = None
    netAmount: Union[str, None] = None
    netAmountInCompanyCurrency: Union[str, None] = None
    unitPrice: Union[str, None] = None
    unitPriceInCompanyCurrency: Union[str, None] = None
    manualUnitCost: Union[bool, None] = None
    taxId: Union[str, None] = None
    unitCost: Union[str, None] = None
    unitCostInCompanyCurrency: Union[str, None] = None

    excluded_keys: Set[str] = {
        "version",
        "grossAmount",
        "grossAmountInCompanyCurrency",
        "netAmount",
        "netAmountInCompanyCurrency",
        "unitPriceInCompanyCurrency",
        "unitCostInCompanyCurrency",
    }


class StatusHistory(Blueprint):
    status: Union[str, None] = None
    statusDate: Union[int, None] = None
    userId: Union[str, None] = None


class SalesInvoice(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    customAttributes: List[WeclappMetaData] = []
    commercialLanguage: Union[str, None] = None
    creatorId: Union[str, None] = None
    description: Union[str, None] = None
    disableRecordEmailingRule: Union[bool, None] = None
    recordComment: Union[str, None] = None
    recordFreeText: Union[str, None] = None
    recordOpening: Union[str, None] = None
    sentToRecipient: Union[bool, None] = None
    tags: list = []
    currencyConversionDate: Union[int, None] = None
    currencyConversionLocked: Union[bool, None] = None
    currencyConversionRate: Union[str, None] = None
    grossAmount: Union[str, None] = None
    grossAmountInCompanyCurrency: Union[str, None] = None
    headerDiscount: Union[str, None] = None
    headerSurcharge: Union[str, None] = None
    netAmount: Union[str, None] = None
    netAmountInCompanyCurrency: Union[str, None] = None
    nonStandardTaxId: Union[str, None] = None
    paymentMethodId: Union[str, None] = None
    recordCurrencyId: Union[str, None] = None
    termOfPaymentId: Union[str, None] = None
    commission: Union[str, None] = None
    commissionSalesPartners: List[CommissionSalesPartners] = []
    customerId: Union[str, None] = None
    dispatchCountryCode: Union[str, None] = None
    factoring: Union[bool, None] = None
    pricingDate: Union[int, None] = None
    responsibleUserId: Union[str, None] = None
    salesChannel: Union[str, None] = None
    servicePeriodFrom: Union[int, None] = None
    servicePeriodTo: Union[int, None] = None
    shipmentMethodId: Union[str, None] = None
    bookingDate: Union[int, None] = None
    bookingText: Union[str, None] = None
    cancellationDate: Union[int, None] = None
    cancellationNumber: Union[str, None] = None
    cancellationSlipCommissionBlock: Union[bool, None] = None
    cancellationSlipCommissionSettlementDone: Union[bool, None] = None
    collectiveInvoicePositionPrintType: Union[str, None] = None
    commissionBlock: Union[bool, None] = None
    commissionSettlementDone: Union[bool, None] = None
    costCenterId: Union[str, None] = None
    costTypeId: Union[str, None] = None
    creditResetsOrderState: Union[bool, None] = None
    customerHabitualExporterLetterOfIntentId: Union[str, None] = None
    deliveryAddress: DeliveryAddress = None
    deliveryDate: Union[int, None] = None
    directDebitFileCreated: Union[bool, None] = None
    directDebitFileLatestDate: Union[int, None] = None
    dueDate: Union[int, None] = None
    dunningBlockDateUntilDate: Union[int, None] = None
    dunningBlockNote: Union[str, None] = None
    dunningBlockState: Union[str, None] = None
    invoiceDate: Union[int, None] = None
    invoiceNumber: Union[str, None] = None
    orderNumberAtCustomer: Union[str, None] = None
    paid: Union[bool, None] = None
    paymentStatus: Union[str, None] = None
    precedingSalesInvoiceId: Union[str, None] = None
    recordAddress: RecordAddress = RecordAddress()
    recordCommentInheritance: Union[bool, None] = None
    recordEmailAddresses: RecordEmailAddresses = RecordEmailAddresses()
    recordFreeTextInheritance: Union[bool, None] = None
    recordOpeningInheritance: Union[bool, None] = None
    salesInvoiceItems: List[SalesInvoiceItems] = []
    salesInvoiceType: Union[str, None] = None
    salesOrderId: Union[str, None] = None
    salesOrders: List[SalesOrders] = []
    sepaDirectDebitMandateId: Union[str, None] = None
    shippingCostItems: List[ShippingCostItems] = []
    shippingDate: Union[int, None] = None
    status: Union[str, None] = None
    statusHistory: List[StatusHistory] = []
    vatRegistrationNumber: Union[str, None] = None

    excluded_keys: Set[str] = {
        "paid",
        "unitCostInCompanyCurrency",
        "unitPriceInCompanyCurrency"
    }



"""This module contains:
- the Weclapp class, which provides methods to interact
with the Weclapp API. It includes methods for sending GET, PUT, POST, and DELETE
requests, as well as an iterator for paginated results.
- the WeclappError class, which is a custom exception class for handling errors
from the Weclapp API.
"""

import json
import os
import logging
from typing import Union, Generator
import requests
from . import config


class WeclappError(Exception):
    """Custom exception class for Weclapp API errors."""

    def __init__(
        self, error_response: requests.Response, api_version: str = config.API_VERSION
    ):
        self.response = error_response
        self.url = str(self.response.url).split(f"/api/{api_version}", maxsplit=1)[-1]
        self.is_json = False
        try:
            error_response = self.response.json()
            self.detail = error_response.get("detail", "")
            self.error = error_response.get("error", "")
            self.status = error_response.get("status", 400)
            self.title = error_response.get("title", "")
            self.error_type = error_response.get("type", "")
            self.validation_errors = error_response.get("validationErrors", [])
            self.messages = error_response.get("messages", [])
            self.is_json = True

        except (json.JSONDecodeError, requests.exceptions.JSONDecodeError):
            self.detail = error_response.text

        super().__init__(self.detail)

    @property
    def is_optimistic_lock(self) -> bool:
        """Checks if the error is an optimistic lock error."""
        if self.detail == config.OPTIMISTIC_LOCK_IDENTIFIER:
            logging.error("Optimistic lock error found.")
            return True
        return False

    def __str__(self) -> str:
        """Returns a string representation of the WeclappError."""
        if not self.is_json:
            return (
                f"{self.response.status_code}: {self.detail} when sending "
                f"request to {self.url}"
            )

        error_message = (
            f"Error {self.title} with code {self.response.status_code}: "
            f"{self.error}, when sending request to {self.url}."
        )
        for message in self.messages:
            if isinstance(message, dict):
                error_message += (
                    f"\n\t\t{message.get('severity')}={message.get('message')};"
                )
        for message in self.validation_errors:
            error_message += f"\n\t\t{message};"
        return error_message


class Weclapp:
    """A class to interact with the Weclapp API.

    Attributes:
        api_token (str): The API token for authentication. If not provided, it will be
            fetched from the environment variable "WECLAPP_API_TOKEN".
        domain (str): The Weclapp domain to connect to. If not provided, it will be
            fetched from the environment variable "WECLAPP_DOMAIN".
        base_url (str): The base URL for the Weclapp API. Generated from domain
            and API version.
        headers (dict): The headers to use for API requests. Generated from the
            authentication token and content type.

    Methods:
        get: Sends a GET request to the Weclapp API.
        put: Sends a PUT request to update an entity in the Weclapp API.
        post: Sends a POST request to create or update an entity in the Weclapp API.
        delete: Sends a DELETE request to remove an entity from the Weclapp API.
        iterator: Yields entities from the Weclapp API based on the provided query.
    """

    def __init__(
        self,
        api_token: str = None,
        domain: str = None,
        api_version: str = config.API_VERSION,
    ):
        if not api_token:
            api_token = os.environ.get(config.API_TOKEN_ENV_VAR)
        if not domain:
            domain = os.environ.get(config.DOMAIN_ENV_VAR)
        domain = self._check_domain(domain)
        self.base_url = f"https://{domain}/webapp/api/{api_version}"
        self.headers = self._get_headers(api_token)

    def _get_headers(self, api_token: str) -> dict:
        """Returns the headers for the weclapp API request."""
        if not api_token:
            raise ValueError(
                "Authentication token for weclapp must be provided as a non-empty string."
            )

        if not isinstance(api_token, str):
            raise TypeError(
                f"Authentication token for weclapp must be a string, but is {type(api_token)}."
            )

        return {
            config.AUTHENTICATION_TOKEN_NAME: api_token,
            "Content-Type": config.DEFAULT_CONTENT_TYPE,
        }

    def _check_domain(self, url: str) -> str:
        """Checks if the provided URL is a valid Weclapp domain."""
        if not url:
            raise ValueError("Weclapp domain must be provided as a non-empty string.")
        url = str(url).strip()
        if not url.endswith(config.WECLAPP_DOMAIN_ENDING):
            raise ValueError(
                f"Weclapp domain {url} must end with {config.WECLAPP_DOMAIN_ENDING}."
            )
        return url

    def _parse_response(
        self,
        response: requests.Response,
        as_type: Union[dict, bytes, list] = dict,
        return_full_result: bool = False,
    ) -> Union[dict, list, bytes, int]:
        """
        Parses the response from the Weclapp API and returns it in the specified format.

        Args:
            response (requests.Response): The response object from the Weclapp API.
            as_type (Union[dict, bytes, list]): The type to return the parsed response as.
            return_full_result (bool): If True, returns the full result including metadata.
                If set to True, the function will always return a dict.

        Returns:
            result (Union[dict, list, bytes, int]): The parsed response in the specified format.
        """
        if not response.ok:
            raise WeclappError(response)

        if as_type == bytes:
            return response.content

        json_response = response.json()

        if return_full_result is True:
            if as_type != dict:
                logging.warning(
                    "If return_full_result is True, function will always return a dict."
                )
            return json_response

        result = (
            json_response[config.DEFAULT_RESPONSE_CONTAINER]
            if config.DEFAULT_RESPONSE_CONTAINER in json_response
            else json_response
        )

        if config.COUNT_REQUEST_IDENTIFIER in response.url:
            if isinstance(result, int):
                return result
        elif as_type == dict:
            if not isinstance(result, dict):
                logging.info(
                    "Not a dict object was returned by Weclapp -> turned it into dict"
                )
                result = {config.DEFAULT_RESPONSE_CONTAINER: result}
            return result
        elif as_type == list:
            if not isinstance(result, list):
                logging.info(
                    "Not a list object was returned by Weclapp -> turned it into list"
                )
                result = [result]
            return result
        else:
            raise TypeError("as_type must be one of dict, list, bytes")

        return result

    def get(
        self,
        entity_name: str,
        entity_id: str = None,
        query: dict = None,
        as_type: Union[dict, bytes, list] = dict,
        include_result: bool = False,
    ) -> Union[dict, list, bytes, int]:
        """Sends a GET request to the Weclapp API.
        Arguments:
            entity_name (str): The name of the entity to retrieve.
            entity_id (str, optional): The ID of the specific entity to retrieve.
            query (dict, optional): Additional query parameters for the request.
            as_type (Union[dict, bytes, list], optional): The type to return the
                parsed response as.
            include_result (bool, optional): If True, returns the full result
                including metadata.
                In this case, the function will always return a dict.
        Returns:
            result (Union[dict, list, bytes, int]): The parsed response in the
                specified format.
        """
        query = query or {}
        if not isinstance(query, dict):
            raise TypeError(f"query must be a dict, but got {type(query).__name__}")

        url = f"{self.base_url}/{entity_name}"

        if entity_id is not None:
            url += f"/id/{entity_id}"

        response = requests.get(
            url,
            headers=self.headers,
            params=query,
            timeout=config.REQUEST_TIMEOUT,
        )
        return self._parse_response(
            response, as_type=as_type, return_full_result=include_result
        )

    def put(
        self,
        entity_name: str,
        entity_id: str,
        body: dict,
        ignore_missing_properties: bool = True,
        query: dict = None,
    ) -> dict:
        """Sends a PUT request to the weclapp API.
        Arguments:
            entity_name (str): The name of the entity to update.
            entity_id (str): The ID of the specific entity to update.
            body (dict): The data to update the entity with.
            ignore_missing_properties (bool): If True, the weclapp API will
                ignore missing properties in the body. Be careful when modifying
                list items, all other (unmodified) items need to be mentioned as
                well or they will be deleted!
            query (dict, optional): Additional query parameters for the request.
        Returns:
            result (dict): The parsed response from the API.
        """
        query = query or {}
        url = f"{self.base_url}/{entity_name}/id/{entity_id}"

        if not isinstance(query, dict):
            raise TypeError(f"query must be a dict, but got {type(query).__name__}")
        if not isinstance(body, dict):
            raise TypeError(f"body must be a dict, but got {type(body).__name__}")

        if ignore_missing_properties is True:
            query.update({"ignoreMissingProperties": True})

        response = requests.put(
            url,
            headers=self.headers,
            data=json.dumps(body),
            params=query,
            timeout=config.REQUEST_TIMEOUT,
        )

        return self._parse_response(response, as_type=dict)

    def post(
        self,
        entity_name: str,
        entity_id: str = None,
        body: dict = None,
        query: dict = None,
    ) -> dict:
        """Sends a POST request to the Weclapp API.
        Arguments:
            entity_name (str): The name of the entity to create or update.
            entity_id (str, optional): The ID of the specific entity to update.
                If None, a new entity will be created.
            body (dict, optional): The data to send in the request body.
            query (dict, optional): Additional query parameters for the request.
        Returns:
            result (dict): The parsed response from the API.
        """
        query = query or {}
        body = body or {}
        if not isinstance(query, dict):
            raise TypeError("query must be a dict")

        if isinstance(body, dict):
            body = json.dumps(body)
        elif isinstance(body, bytes) and entity_name in config.BYTE_TYPE_BODY_ENTITIES:
            logging.info("Sending bytes type body for post request.")
        else:
            raise TypeError("body must be a dict or bytes, when uploading document")

        url = f"{self.base_url}/{entity_name}"
        if entity_id is not None:
            url += f"/id/{entity_id}"

        response = requests.post(
            url,
            headers=self.headers,
            data=body,
            params=query,
            timeout=config.REQUEST_TIMEOUT,
        )

        return self._parse_response(response=response, as_type=dict)

    def delete(
        self,
        entity_name: str,
        entity_id: str,
    ) -> None:
        """Sends a DELETE request to the Weclapp API to delete an entity.
        Arguments:
            entity_name (str): The name of the entity to delete.
            entity_id (str): The ID of the specific entity to delete.
        """

        logging.warning("Deleting entity %s with id %s", entity_name, entity_id)
        url = f"{self.base_url}/{entity_name}/id/{entity_id}"
        response = requests.delete(
            url, headers=self.headers, timeout=config.REQUEST_TIMEOUT
        )

        if not response.ok:
            raise WeclappError(response)

    def _check_iteration_parameters(
        self,
        query: dict,
        page_size: int,
        start_page: int,
        max_entities: int,
    ) -> dict:
        """Checks and prepares the parameters for iteration over entities. If any
        of the parameters are of the wrong type or out of range, it raises an error.
        Returns:
            query (dict): The updated query with pagination parameters.
        """
        query = query or {}
        if not isinstance(query, dict):
            raise TypeError(f"query must be a dict, but got {type(query).__name__}")

        if not isinstance(page_size, int) or not 0 < page_size <= 1000:
            raise ValueError(
                f"Page size must be an integer 0 < page_size <= 1000, but is "
                f"{page_size} of type {type(page_size).__name__}"
            )

        if not isinstance(start_page, int) or start_page < 1:
            raise ValueError(
                f"Start page must be an integer >= 1, but is {start_page} of "
                f"type {type(start_page).__name__}"
            )

        if max_entities and (not isinstance(max_entities, int) or max_entities < 1):
            raise ValueError(
                f"max_entities must be an integer >= 1, but is {max_entities} "
                f"of type {type(max_entities).__name__}"
            )

        if "pageSize" not in query:
            query["pageSize"] = page_size
        elif "pageSize" in query and page_size != query["pageSize"]:
            logging.warning(
                "pageSize in query (%s) does not match the page_size argument "
                "(%s). Using pageSize from query.",
                query["pageSize"],
                page_size,
            )
        return query

    def iterator(
        self,
        entity_name: str,
        query: dict = None,
        page_size: int = 100,
        start_page: int = 1,
        max_entities: int = None,
        enable_logging: bool = True,
    ) -> Generator[dict, None, None]:
        """Yields all items (dict) of an entityName, that satisfy the query.
        Arguments:
            entity_name (str): The name of the entity to iterate over.
            query (dict, optional): Additional query parameters for the request.
            page_size (int, optional): The number of items to retrieve per page.
                Default is 100, maximum is 1000.
            start_page (int, optional): The page number to start from. Default is 1.
            max_entities (int, optional): Maximum number of entities to yield.
                If None, all entities will be yielded.
            enable_logging (bool, optional): If True, enables logging of the
                iteration process.
        Yields:
            dict: The next entity in the iteration.
        """
        processed_items = 0
        page = start_page
        query = self._check_iteration_parameters(
            query, page_size, start_page, max_entities
        )

        if enable_logging:
            logging.warning("--- Iteration for %s started ---", entity_name)

        while True:
            query["page"] = page
            entity_list = self.get(
                entity_name=entity_name,
                query=query,
                as_type=list,
            )

            if not isinstance(entity_list, list):
                raise ValueError(
                    f"Expected a list of entities, but got {type(entity_list).__name__}: "
                    f"{entity_list}"
                )

            entities_found = len(entity_list)
            if entities_found <= 0:
                break

            if enable_logging:
                logging.warning(
                    "--- Page %s: Found %s entities ---", page, entities_found
                )

            for obj in entity_list:
                processed_items += 1
                if max_entities and max_entities <= processed_items:
                    break
                yield obj

            page += 1
            if max_entities and max_entities <= processed_items:
                break
        if enable_logging:
            logging.warning("--- Iteration for %s finished ---", entity_name)



"""Module for generating and managing custom attributes (CAT) for Weclapp entities."""

import logging
import re
import unicodedata
import json
import os
from typing import Tuple, List
from .custom_attribute_class import CustomAttributeClass
from ..weclapp import Weclapp
from . import config


class CustomAttributeGenerator:
    """Generator for custom attributes classes of Weclapp entities.
    Args:
        entity_name (str): The name of the Weclapp entity for which to generate
            custom attributes.
        target_directory (str): The directory where the generated files will
            be saved.
    Usage:
        1. Create an instance of CustomAttributeGenerator with the entity
        name and target directory.
        2. Call the main method to generate the custom attributes.
        3. If necessary, create json file named "special_attributes.json"
        in the target directory to assign names to system attributes
        (attributes that are not created by the user). Do this before
        running the generator.
        4. Import CustomAttributes class from __init__.py to access custom
        attributes of all Weclapp entities. Or import the specific
        custom attribute class for the entity you are interested in.
    """

    def __init__(self, target_directory: str):
        self.target_directory = target_directory
        self.weclapp = Weclapp()
        self.collected_custom_attributes: List[CustomAttributeClass] = []
        self.special_attributes = self.load_special_attributes()

    def load_special_attributes(self) -> dict:
        """Loads the special attributes from a JSON file.
        This file contains mappings of custom attribute IDs to their names.
        If the file does not exist, it returns an empty dictionary.
        """
        os.makedirs(self.target_directory, exist_ok=True)

        if not os.path.exists(f"{self.target_directory}/{config.STATIC_IDS_JSON_FILE}"):
            logging.warning(
                "File %s not found in %s. Provide the file to assign names to "
                "system attributes.",
                config.STATIC_IDS_JSON_FILE,
                self.target_directory,
            )
            return {}

        with open(
            f"{self.target_directory}/{config.STATIC_IDS_JSON_FILE}",
            "r",
            encoding="utf-8",
        ) as f:
            return dict(json.load(f))

    def main(self) -> None:
        """Main method to run the custom attribute generator."""
        self.collect_custom_attributes()
        custom_attributes_json, file_content = self.construct_file_content()
        self.save_files(custom_attributes_json, file_content)

    def collect_custom_attributes(self):
        """Finds the default custom attributes for the given entity and adds them to the set"""
        custom_attributes_definitions = self.retrieve_custom_attributes()
        for attribute_dict in custom_attributes_definitions:
            cat = self.setup_custom_attribute_definition(attribute_dict)
            self.collected_custom_attributes.append(cat)

    def retrieve_custom_attributes(self) -> Weclapp:
        """Returns a list of all active custom attribute definitions from Weclapp."""
        attributes = []
        for custom_attribute in self.weclapp.iterator(
            entity_name="customAttributeDefinition",
            query=config.ATTRIBUTE_QUERY,
            page_size=1000,
        ):
            attributes.append(custom_attribute)

        return attributes

    def setup_custom_attribute_definition(
        self, attribute_dict: dict
    ) -> CustomAttributeClass:
        """Fetches the custom attribute definition from Weclapp, transforms the
        data to Python friendly format and returns a CustomAttributeClass instance."""
        attr_key = (
            attribute_dict["attributeKey"]
            if attribute_dict["id"] not in self.special_attributes
            else self.special_attributes[attribute_dict["id"]]
        )
        attr_key = self.to_python_variable(attr_key)
        attr_type = attribute_dict["attributeType"]
        group_name = attribute_dict.get("groupName", config.DEFAULT_GROUP_NAME).strip()

        selectable_values = {}
        if "selectableValues" in attribute_dict and attribute_dict["selectableValues"]:
            for value in attribute_dict["selectableValues"]:
                name = value["value"].replace("___", "_").strip()
                match = re.search(r"\((.*?)\)", name)
                if match:
                    name = match.group(1)
                name = self.to_python_variable(name)
                selectable_values[name] = value["id"]

        return CustomAttributeClass(
            group_name=group_name,
            attr_key=attr_key,
            attr_id=attribute_dict["id"],
            attr_type=attr_type,
            selectable_values=selectable_values,
        )

    def to_python_variable(self, raw_string: str) -> str:
        """Converts a raw string to a Python variable name."""
        # Normalize unicode (e.g., ä -> a)
        s = (
            unicodedata.normalize("NFKD", raw_string)
            .encode("ascii", "ignore")
            .decode("ascii")
        )
        # Replace spaces and slashes with underscores
        s = re.sub(r"[ /]", "_", s)
        # Remove any character that is not a letter, number, or underscore
        s = re.sub(r"\W", "", s)
        # Ensure it doesn't start with a number
        if s and s[0].isdigit():
            s = "_" + s
        return s.lower()

    def construct_file_content(self) -> Tuple[str, dict]:
        """Constructs the python file content and the JSON data for the custom attributes."""
        custom_attributes_json = {}
        file_content = (
            '"""Automatically generated file for Weclapp custom attributes."""\n\n'
            "import json\n"
            "from types import SimpleNamespace\n\n\n"
            "class CustomAttributes:\n"
        )
        file_content += '    """Class to handle Weclapp custom attributes."""\n'
        file_content += (
            "    def __init__(self, data: dict = None):\n"
            "        if data is None:\n"
            "            with open(\n"
            f'                "{self.target_directory}/{config.JSON_DATA_FILE_NAME}'
            '.json", "r", encoding="utf-8"\n'
            f"            ) as file:\n"
            f"                data = json.load(file)\n\n"
            f"        for key, value in data.items():\n"
            f"            setattr(self, key, SimpleNamespace(**value))\n"
        )

        previous_group_name = None
        for custom_attribute in sorted(
            self.collected_custom_attributes, key=lambda x: x.group_name
        ):
            if previous_group_name != custom_attribute.group_name:
                previous_group_name = custom_attribute.group_name
                file_content += "\n"
                file_content += f"        # {custom_attribute.group_name}\n"

            custom_attributes_json[custom_attribute.attr_key] = (
                custom_attribute.to_dict()
            )
            file_content += custom_attribute.generate_python_line()

        return custom_attributes_json, file_content

    def save_files(self, custom_attributes_json: dict, file_content: str) -> None:
        """Saves the generated custom attributes JSON and Python file to the
        target directory."""
        os.makedirs(
            f"{self.target_directory}/{config.JSON_DATA_FILE_NAME}", exist_ok=True
        )

        with open(
            f"{self.target_directory}/{config.JSON_DATA_FILE_NAME}.json",
            "w+",
            encoding="utf-8",
        ) as f:
            json.dump(custom_attributes_json, f, indent=4)

        with open(
            f"{self.target_directory}/__init__.py",
            "w+",
            encoding="utf-8",
        ) as f:
            f.write(file_content)



"""This module provides functions for converting weclapp timestamps
to datetime objects and for converting datetime objects to various
string or integer representations."""

import datetime
from typing import Union
import zoneinfo


def convert_to_datetime(
    timestamp: Union[str, int, float], timezone: str = "Europe/Berlin"
) -> datetime.datetime:
    """Convert a weclapp timestamp to a datetime object in the specified timezone."""
    if not timestamp:
        raise ValueError("Timestamp must be provided as a non-empty string or number.")
    if isinstance(timestamp, str):
        try:
            float(timestamp)
        except ValueError:
            raise ValueError(
                f"Invalid time format: {timestamp}. Must be a numeric string or a number."
            )
    if not isinstance(timestamp, (int, float, str)):
        raise TypeError(
            f"Invalid type for timestamp: {type(timestamp)}. Must be int, float, or str."
        )
    try:
        timestamp = float(timestamp) / 1000
        relevant_time = datetime.datetime.fromtimestamp(
            timestamp, tz=zoneinfo.ZoneInfo(timezone)
        )

        return relevant_time
    except OverflowError as of:
        raise OverflowError(
            f"{of}, while converting {timestamp} to datetime object"
        ) from of


def convert_to_time_format(
    input_time: Union[datetime.date, datetime.datetime],
    conversion_format: str = "weclapp",
    timezone: str = "Europe/Berlin",
) -> Union[str, int]:
    """Converts a datetime object to a string or integer representation.
    Args:
        input_time (Union[datetime.date, datetime.datetime]): The
            time to convert. Can be a datetime object or date object.
        conversion_format (str): The format to convert the time to. Must be one of
            'unix', 'weclapp', 'emailDate', 'utcDate', or 'ads'.
            - 'unix': Returns the time as a Unix timestamp (int).
            - 'weclapp': Returns the time as a weclapp timestamp (int).
            - 'emailDate': Returns the time as an email date string (str).
            - 'utcDate': Returns the time as a UTC date string (str).
            - 'ads': Returns the time as an ADS date string (str).
    Returns:
        Union[str, int]: The converted time in the specified format.
    """
    try:
        if conversion_format == "unix":
            return int(input_time.timestamp())
        if conversion_format == "weclapp":
            return int(input_time.timestamp()) * 1000
        if conversion_format == "emailDate":
            return str(input_time.strftime("%d.%m.%Y"))
        if conversion_format == "utcDate":
            return str(input_time.strftime("%Y-%m-%d"))
        if conversion_format == "ads":
            # 2022-12-04 11:11:11-01:00
            aware_time = datetime.datetime(
                input_time.year,
                input_time.month,
                input_time.day,
                input_time.hour,
                input_time.minute,
                input_time.second,
                tzinfo=zoneinfo.ZoneInfo(timezone),
            )
            return aware_time.isoformat().replace("T", " ")

        raise ValueError(
            f"Invalid conversion format: {conversion_format}. Must be one of "
            "'unix', 'weclapp', 'emailDate', 'utcDate', or 'ads'."
        )

    except OverflowError as of:
        raise OverflowError(
            f"{of}, while transforming {input_time} to {conversion_format}"
        ) from of



"""Configuration for weclapp_classes.blueprint package. """

from typing import Literal

SALES_ORDER = "salesOrder"

CUSTOM_ATTRIBUTE_TYPES = [
    "selectedValueId",
    "stringValue",
    "booleanValue",
    "dateValue",
    "selectedValues",
    "entityId",
    "entityReferences",
    "numberValue",
]
CUSTOM_ATTRIBUTE_TYPES_LITERAL = Literal[
    "selectedValueId",
    "stringValue",
    "booleanValue",
    "dateValue",
    "selectedValues",
    "entityId",
    "entityReferences",
    "numberValue",
]
LIST_CUSTOM_ATTRIBUTE_TYPES = ["selectedValues", "entityReferences"]
LIST_CUSTOM_ATTRIBUTE_TYPES_LITERAL = Literal["selectedValues", "entityReferences"]
SINGLE_VALUE_CUSTOM_ATTRIBUTE_TYPES = [
    "selectedValueId",
    "stringValue",
    "booleanValue",
    "dateValue",
    "entityId",
    "numberValue",
]
SINGLE_VALUE_CUSTOM_ATTRIBUTE_TYPES_LITERAL = Literal[
    "selectedValueId",
    "stringValue",
    "booleanValue",
    "dateValue",
    "entityId",
    "numberValue",
]
INT_CUSTOM_ATTRIBUTE_TYPES = ["selectedValueId", "entityId", "dateValue"]
FLOAT_CUSTOM_ATTRIBUTE_TYPES = ["numberValue"]


# --------- Weclapp Class Creator ---------
CREATION_QUERY = {"serializeNulls": "", "sort": "-createdDate"}

STATIC_IMPORTS_MODEL_FILES = (
    '"""This code was dynamically created using WeclappClassCreator from pyweclapp"""'
    "\n\n"
    "from typing import Union, Optional, List\n"
    "from .blueprints import Blueprint, WeclappMetaData\n\n\n"
)
INIT_FILE_NAME = "__init__.py"
STATIC_IMPORTS_INIT = "from .blueprints.custom_attributes_model import WeclappMetaData"
INIT_FILE_DOC_STRING = (
    '"""Module for weclapp classes package imports. Add more classes if necessary."""\n'
)

EXCLUDED_KEYS = {
    "used_attributes",
    "used_keys",
    "excluded_keys",
    "createdDate",
    "lastModifiedDate",
    "statusHistory",
    "lastModifiedByUserId"
}



"""This code was dynamically created using WeclappClassCreator from pyweclapp"""

from typing import Union, List
from .blueprints import Blueprint, WeclappMetaData


class InvoiceAddress(Blueprint):
    city: Union[str, None] = None
    company: Union[str, None] = None
    company2: Union[str, None] = None
    countryCode: Union[str, None] = None
    firstName: Union[str, None] = None
    globalLocationNumber: Union[str, None] = None
    lastName: Union[str, None] = None
    middleName: Union[str, None] = None
    phoneNumber: Union[str, None] = None
    postOfficeBoxCity: Union[str, None] = None
    postOfficeBoxNumber: Union[str, None] = None
    postOfficeBoxZipCode: Union[str, None] = None
    salutation: Union[str, None] = None
    state: Union[str, None] = None
    street1: Union[str, None] = None
    street2: Union[str, None] = None
    titleId: Union[str, None] = None
    zipcode: Union[str, None] = None


class RecipientAddress(Blueprint):
    city: Union[str, None] = None
    company: Union[str, None] = None
    company2: Union[str, None] = None
    countryCode: Union[str, None] = None
    firstName: Union[str, None] = None
    globalLocationNumber: Union[str, None] = None
    lastName: Union[str, None] = None
    middleName: Union[str, None] = None
    phoneNumber: Union[str, None] = None
    postOfficeBoxCity: Union[str, None] = None
    postOfficeBoxNumber: Union[str, None] = None
    postOfficeBoxZipCode: Union[str, None] = None
    salutation: Union[str, None] = None
    state: Union[str, None] = None
    street1: Union[str, None] = None
    street2: Union[str, None] = None
    titleId: Union[str, None] = None
    zipcode: Union[str, None] = None


class SalesOrders(Blueprint):
    id: Union[str, None] = None


class StatusHistory(Blueprint):
    status: Union[str, None] = None
    statusDate: Union[int, None] = None
    userId: Union[str, None] = None


class CustomerDeliveryAddress(Blueprint):
    city: Union[str, None] = None
    company: Union[str, None] = None
    company2: Union[str, None] = None
    countryCode: Union[str, None] = None
    firstName: Union[str, None] = None
    globalLocationNumber: Union[str, None] = None
    lastName: Union[str, None] = None
    middleName: Union[str, None] = None
    phoneNumber: Union[str, None] = None
    postOfficeBoxCity: Union[str, None] = None
    postOfficeBoxNumber: Union[str, None] = None
    postOfficeBoxZipCode: Union[str, None] = None
    salutation: Union[str, None] = None
    state: Union[str, None] = None
    street1: Union[str, None] = None
    street2: Union[str, None] = None
    titleId: Union[str, None] = None
    zipcode: Union[str, None] = None


class CustomerInvoiceAddress(Blueprint):
    city: Union[str, None] = None
    company: Union[str, None] = None
    company2: Union[str, None] = None
    countryCode: Union[str, None] = None
    firstName: Union[str, None] = None
    globalLocationNumber: Union[str, None] = None
    lastName: Union[str, None] = None
    middleName: Union[str, None] = None
    phoneNumber: Union[str, None] = None
    postOfficeBoxCity: Union[str, None] = None
    postOfficeBoxNumber: Union[str, None] = None
    postOfficeBoxZipCode: Union[str, None] = None
    salutation: Union[str, None] = None
    state: Union[str, None] = None
    street1: Union[str, None] = None
    street2: Union[str, None] = None
    titleId: Union[str, None] = None
    zipcode: Union[str, None] = None


class IncomingGoodsItems(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    customAttributes: List[WeclappMetaData] = []
    articleId: Union[str, None] = None
    note: Union[str, None] = None
    positionNumber: Union[int, None] = None
    quantity: Union[str, None] = None
    description: Union[str, None] = None
    descriptionFixed: Union[bool, None] = None
    itemType: Union[str, None] = None
    manualQuantity: Union[bool, None] = None
    parentItemId: Union[str, None] = None
    title: Union[str, None] = None
    unitId: Union[str, None] = None
    dropshippingShipmentItemId: Union[str, None] = None
    itemGroup: Union[str, None] = None
    purchaseOrderItemId: Union[str, None] = None
    returnAssessmentId: Union[str, None] = None
    returnDescription: Union[str, None] = None
    returnErrorId: Union[str, None] = None
    returnReasonId: Union[str, None] = None
    returnRectificationId: Union[str, None] = None
    salesOrderItemId: Union[str, None] = None


class PurchaseOrders(Blueprint):
    id: Union[str, None] = None


class ReturnAddress(Blueprint):
    city: Union[str, None] = None
    company: Union[str, None] = None
    company2: Union[str, None] = None
    countryCode: Union[str, None] = None
    firstName: Union[str, None] = None
    globalLocationNumber: Union[str, None] = None
    lastName: Union[str, None] = None
    middleName: Union[str, None] = None
    phoneNumber: Union[str, None] = None
    postOfficeBoxCity: Union[str, None] = None
    postOfficeBoxNumber: Union[str, None] = None
    postOfficeBoxZipCode: Union[str, None] = None
    salutation: Union[str, None] = None
    state: Union[str, None] = None
    street1: Union[str, None] = None
    street2: Union[str, None] = None
    titleId: Union[str, None] = None
    zipcode: Union[str, None] = None


class IncomingGoods(Blueprint):
    id: Union[str, None] = None
    createdDate: Union[int, None] = None
    lastModifiedDate: Union[int, None] = None
    version: Union[str, None] = None
    customAttributes: List[WeclappMetaData] = []
    commercialLanguage: Union[str, None] = None
    creatorId: Union[str, None] = None
    description: Union[str, None] = None
    disableRecordEmailingRule: Union[bool, None] = None
    recordComment: Union[str, None] = None
    recordFreeText: Union[str, None] = None
    recordOpening: Union[str, None] = None
    sentToRecipient: Union[bool, None] = None
    tags: list = []
    invoiceAddress: InvoiceAddress = InvoiceAddress()
    recipientAddress: RecipientAddress = RecipientAddress()
    salesOrders: List[SalesOrders] = []
    status: Union[str, None] = None
    statusHistory: List[StatusHistory] = []
    customerDeliveryAddress: CustomerDeliveryAddress = CustomerDeliveryAddress()
    customerInvoiceAddress: CustomerInvoiceAddress = CustomerInvoiceAddress()
    deliveryNoteNumber: Union[str, None] = None
    dhlReceiverId: Union[str, None] = None
    dropshippingShipmentId: Union[str, None] = None
    incomingGoodsItems: List[IncomingGoodsItems] = []
    incomingGoodsNumber: Union[str, None] = None
    incomingGoodsType: Union[str, None] = None
    invoiceRecipientId: Union[str, None] = None
    purchaseOrders: List[PurchaseOrders] = []
    relatedShipmentId: Union[str, None] = None
    responsibleUserId: Union[str, None] = None
    returnAddress: ReturnAddress = ReturnAddress()
    senderCustomerNumber: Union[str, None] = None
    senderPartyId: Union[str, None] = None
    senderSupplierNumber: Union[str, None] = None
    shippingReturnCarrierId: Union[str, None] = None
    sourceWarehouseId: Union[str, None] = None
    warehouseId: Union[str, None] = None



"""This module provides the DocManager class to manage documents for a specific
entity in Weclapp."""

from typing import List
from ..weclapp import Weclapp
from .document import Document


class DocManager:
    """A class to manage documents for a specific entity in Weclapp."""

    def __init__(self, entity_name: str, entity_id: str) -> None:
        self.entity_name = entity_name
        self.entity_id = entity_id
        self.weclapp = Weclapp()
        self.documents: List[Document] = []

    def get_documents(self) -> List[Document]:
        """Queries all documents of the entity and returns them as a list of
        Document objects"""
        response = self.weclapp.get(
            entity_name="document",
            query={"entityName": self.entity_name, "entityId": self.entity_id},
            as_type=list
        )
        for doc in response:
            self.documents.append(Document(**doc))

        return self.documents



"""Configuration settings for Weclapp API interactions."""

API_TOKEN_ENV_VAR = "WECLAPP_API_TOKEN"
DOMAIN_ENV_VAR = "WECLAPP_DOMAIN"

API_VERSION = "v2"
WECLAPP_DOMAIN_ENDING = ".weclapp.com"
AUTHENTICATION_TOKEN_NAME = "AuthenticationToken"
DEFAULT_CONTENT_TYPE = "application/json"
REQUEST_TIMEOUT = 120  # seconds

# ----------------- Response Processing -----------------
DEFAULT_RESPONSE_CONTAINER = "result"
COUNT_REQUEST_IDENTIFIER = "/count"
OPTIMISTIC_LOCK_IDENTIFIER = "optimistic lock error"

# ----------------- Entity Specifics -----------------
BYTE_TYPE_BODY_ENTITIES = ["document"]



"""Setup script for the pyweclapp package."""

from setuptools import find_packages, setup

try:
    with open("src/README.md", "r", encoding="utf-8") as f:
        LONG_DESCRIPTION = f.read()
except FileNotFoundError:
    LONG_DESCRIPTION = (
        "Provides methods, classes and classbuilders to interact with the Weclapp API"
    )

try:
    with open("requirements.txt", "r", encoding="utf-8") as f:
        REQUIRED = f.read().splitlines()
except FileNotFoundError:
    REQUIRED = [
        "requests>=2.28.2",
        "pydantic>=1.10.4",
        "jsonschema>=4.17.3",
        "pytz>=2022.7.1",
    ]

setup(
    name="pyweclapp",
    version="0.1.12",
    description="Provides methods, classes and classbuilders to interact with the Weclapp API",
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/AltruanGmbH/pyweclapp",
    author="Altruan GmbH",
    author_email="admin@altruan.de",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    install_requires=REQUIRED,
    extras_require={
        "dev": ["twine>=4.0.2"],
    },
    python_requires=">=3.9",
)



"""Core module for Blueprint class, which serves as a base class for Weclapp entities."""

import logging
import re
from typing import Any, Union, Literal, Optional, List, get_args, Set
from types import SimpleNamespace
from pydantic import BaseModel
from ...weclapp import Weclapp
from .custom_attributes_model import WeclappMetaData
from . import config


class UpdateSettings:
    """Class to hold settings for building update dictionaries.
    Args:
        update_type (str): Type of update. Can be 'full' or 'used'.
        include_version (bool): Whether to include the version in the update.
        creation_mode (bool): Whether the update is for creation mode.
    """

    def __init__(
        self,
        update_type: str,
        include_version: bool,
        creation_mode: bool,
        excluded_keys: Optional[Set[str]] = None,
    ):
        self.update_type: str = update_type
        self.include_version: bool = include_version
        self.creation_mode: bool = creation_mode
        self.excluded_keys: Optional[Set[str]] = excluded_keys
        if include_version is False:
            self.excluded_keys.add("version")
        if creation_mode:
            self.excluded_keys.update(["id", "version"])
            self.update_type = "full"


class Blueprint(BaseModel):
    """Base class for Weclapp entities, providing methods for handling custom
    attributes and entity updates."""

    used_attributes: Union[dict, None] = {}

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.used_attributes = {}

    @classmethod
    def from_weclapp(cls, entity_id: str):
        """Gets an entity from weclapp by its id and returns a new instance of
        the corresponding class with the attributes set to the values."""
        if not entity_id:
            raise AssertionError(
                "Entity ID must not be None or empty for from_weclapp()"
            )
        entity_name = cls.__name__
        entity_name = entity_name[:1].lower() + entity_name[1:]
        response = Weclapp().get(
            entity_name=entity_name, entity_id=entity_id, as_type=dict
        )
        if not isinstance(response, dict):
            raise AssertionError(
                f"Response from weclapp is not a dict, but a {type(response).__name__}"
            )
        return cls(**response)

    def custom_attribute(
        self, attribute_identifier: Union[str, int, tuple]
    ) -> WeclappMetaData:
        """Iterates through the customAttributes and returns the first item
        where the name matches the attribute_identifier.
        Args:
            attribute_identifier (Union[str, int, tuple]): The identifier of the
                custom attribute. If string or integer is provided, it will be used
                as the name of the custom attribute. If a tuple of custom attribute
                is provided, it will be used to get the id of the custom attribute.
        """
        if hasattr(self, "customAttributes"):
            if isinstance(attribute_identifier, str):
                attribute_id = attribute_identifier
            elif isinstance(attribute_identifier, int):
                attribute_id = str(attribute_identifier)
            elif isinstance(attribute_identifier, SimpleNamespace):
                attribute_id = attribute_identifier.id
            else:
                raise TypeError(
                    f"Value must be a string, an integer, or a SimpleNamespace, "
                    f"not {type(attribute_identifier).__name__}"
                )

            for custom_attribute in self.customAttributes:
                if (
                    isinstance(custom_attribute, WeclappMetaData)
                    and custom_attribute.attributeDefinitionId == attribute_id
                ):
                    return custom_attribute
            return WeclappMetaData(attributeDefinitionId=attribute_id)

        raise KeyError(f"Attribute customAttributes not found in {type(self).__name__}")

    def cat(self, attribute_identifier: Union[str, int, tuple]) -> WeclappMetaData:
        """Returns the custom attribute with the given identifier. Shorter alias
        for custom_attribute.

        Args:
            attribute_identifier (Union[str, int, tuple]): The identifier of the
                custom attribute. If string or integer is provided, it will be used
                as the name of the custom attribute. If a tuple of custom attribute
                is provided, it will be used to get the id of the custom attribute.
        """

        return self.custom_attribute(attribute_identifier=attribute_identifier)

    def add_used_attribute(self, attribute_name: str) -> None:
        """Adds an attribute to the used_attributes dictionary if it exists in
        the instance."""
        if hasattr(self, attribute_name):
            self.used_attributes[attribute_name] = getattr(self, attribute_name)

    def remove_used_attribute(self, attribute_name: str) -> None:
        """Deletes an attribute from the used_attributes dictionary if it exists."""
        if attribute_name in self.used_attributes:
            del self.used_attributes[attribute_name]

    def update_custom_attribute(self, key: str, value: Any) -> None:
        """Updates a custom attribute with the given key and value."""
        if not hasattr(self, key):
            raise KeyError(
                f"{type(self).__name__} has no attribute {key} -> "
                "check spelling of the custom attribute, "
                "make sure it is included in the customAttributes list"
            )
        setattr(self, key, value)

    def add_tag(self, new_tag: str) -> None:
        """Adds a tag to a list of tags. Allowed characters are a-z, 0-9 and
        "-", "-", " " -> others will be replaced by "_"
        """

        if hasattr(self, "tags"):
            new_tag = re.sub(r"[^a-zA-Z0-9\-_ ]", "_", str(new_tag).strip())
            current_tags = self.__dict__.get("tags", [])

            if isinstance(current_tags, list):
                if new_tag not in current_tags:
                    current_tags.append(str(new_tag).strip())
                    self.__dict__["tags"] = current_tags
                    self.add_used_attribute("tags")
                else:
                    logging.info("Tag %s is already in tags!", new_tag)
            else:
                raise TypeError("Tags attribute of the entity is not a list!")
        else:
            raise KeyError("No tags attribute found in this class")

    def replace_tag(self, new_tag: str, tag_identifier: str = None, regex: str = None):
        """Replaces and adds new_tag. Allowed characters are a-z, 0-9 and "-",
        "-", " " -> others will be replaced by "_".
        1. If tag_identifier is provided, it will replace the tag that contains
        the tag_identifier.
        2. If regex is provided, it will replace all tags that match the regex.
        3. If neither is provided, it will add the new_tag to the list of tags.
        """

        if hasattr(self, "tags"):
            current_tags = self.__dict__.get("tags", [])
            if isinstance(current_tags, list):
                tag_value = re.sub(r"[^a-zA-Z0-9\-_ ]", "_", str(new_tag).strip())

                if regex is not None:
                    current_tags_set = {
                        str(tag) for tag in current_tags if not re.match(regex, tag)
                    }
                elif tag_identifier and isinstance(tag_identifier, str):
                    current_tags_set = {
                        str(tag)
                        for tag in current_tags
                        if str(tag_identifier).lower() not in str(tag).lower()
                    }
                else:
                    current_tags_set = set(current_tags)

                current_tags_set.add(tag_value)

                if not all(tag in current_tags_set for tag in current_tags) or len(
                    current_tags
                ) != len(current_tags_set):
                    self.__dict__["tags"] = list(current_tags_set)
                    self.add_used_attribute("tags")
            else:
                raise TypeError("Tags attribute of the entity is not a list!")

        else:
            raise KeyError("No tags attribute found in this class")

    def delete_tag(
        self, tag_to_remove: str = None, regex: str = None, allow_empty: bool = False
    ) -> None:
        """Deletes a tag from the list of tags.
        Args:
            tag_to_remove (str, optional): The tag to remove. If provided, it will
                remove the tag that matches this value.
            regex (str, optional): A regex pattern to match tags to remove. If
                provided, it will remove all tags that match this regex.
            allow_empty (bool, optional): If True, allows the tags list to be
                empty after deletion. If False, will not update the tags list if
                it becomes empty.
        """
        if tag_to_remove is None and regex is None:
            raise AssertionError("Please provide a tag_to_remove or a regex")

        if hasattr(self, "tags"):
            current_tags = self.__dict__.get("tags", [])
            if isinstance(current_tags, list):
                initial_length = len(current_tags)
                if regex is not None:
                    current_tags = [
                        str(tag) for tag in current_tags if not re.match(regex, tag)
                    ]
                else:
                    try:
                        tag_to_remove = re.sub(
                            r"[^a-zA-Z0-9\-_ ]", "_", str(tag_to_remove).strip()
                        )
                        current_tags.remove(tag_to_remove)
                    except ValueError:
                        pass

                if current_tags == [] and not allow_empty:
                    logging.warning("Tags list is empty! -> list will not be updated")
                    return

                if len(current_tags) != initial_length:
                    self.__dict__["tags"] = current_tags
                    self.add_used_attribute("tags")
            else:
                raise TypeError("Tags attribute of the entity is not a list!")
        else:
            raise KeyError("No tags attribute found in this class")

    def build_update_dictionary(
        self,
        update_type: Literal["full", "used"] = "used",
        include_version: bool = False,
        creation_mode: bool = False,
    ) -> dict:
        """Builds a dictionary for updating or creating an entity in Weclapp.
        Args:
            update_type (Literal[full, used]): Type of update. Defaults to 'used'.
                - 'full': Update dictionary will include all attributes of the
                    instance. This type of update should be used with caution
                    as it could potentially overwrite some values to empty if
                    they are not set in the instance.
                - 'used': Update dictionary will include only attributes that
                    have been used.
            include_version (bool): If True, includes the version in the update.
                Be cautious when using this option as it may lead to optimistic
                lock errors. Defaults to False.
            enable_logging (bool): If True, logs the update action. Defaults to False.
        Returns:
            dict: Dictionary representing the entity to be updated or created.
        """
        excluded_keys = config.EXCLUDED_KEYS.copy()
        if hasattr(self, "excluded_keys"):
            excluded_keys.update(self.excluded_keys)
        update_settings = UpdateSettings(
            update_type=update_type,
            include_version=include_version,
            creation_mode=creation_mode,
            excluded_keys=excluded_keys
        )
        logging.warning("Excluded keys for update: %s", update_settings.excluded_keys)

        data_to_send = {}
        for key, value in self.__dict__.items():
            if key in update_settings.excluded_keys:
                continue
            if update_settings.creation_mode and value is None:
                continue
            if key == "customAttributes":
                updated_custom_attributes = self._handle_custom_attributes(
                    value, "full"
                )
                data_to_send[key] = updated_custom_attributes

            elif isinstance(value, list):
                value = self._handle_list_values(key, value, update_settings)
                if value:
                    data_to_send[key] = value

            elif hasattr(value, "build_update_dictionary"):
                object_dictionary = value.build_update_dictionary(
                    update_settings.update_type,
                    include_version=False,
                    creation_mode=update_settings.creation_mode,
                )
                if object_dictionary:
                    data_to_send[key] = object_dictionary

            elif key in self.used_attributes or update_settings.update_type == "full":
                data_to_send[key] = value

        return data_to_send

    def update_entity(
        self,
        update_type: Literal["full", "used"] = "used",
        include_version: bool = False,
        enable_logging: bool = False,
    ) -> dict:
        """Updates the entity in Weclapp with the attributes set in the instance.
        Args:
            update_type (Literal[full, used]): Type of update. Defaults to 'used'.
                - 'full': Update dictionary will include all attributes of the
                    instance. This type of update should be used with caution
                    as it could potentially overwrite some values to empty if
                    they are not set in the instance.
                - 'used': Update dictionary will include only attributes that
                    have been used.
            include_version (bool): If True, includes the version in the update.
                Be cautious when using this option as it may lead to optimistic
                lock errors. Defaults to False.
            enable_logging (bool): If True, logs the update action. Defaults to False.
        """

        body = self.build_update_dictionary(
            update_type=update_type, include_version=include_version
        )
        if len(body) == 0:
            logging.info(
                "No changes detected in %s -> nothing to update", self.__entity_name__
            )
            return self.build_update_dictionary(update_type="full")
        if hasattr(self, "id"):
            updated_entity = Weclapp().put(
                entity_name=self.__entity_name__, entity_id=self.id, body=body
            )
            if enable_logging:
                logging.warning("Updated %s with body: %s", type(self).__name__, body)
            return updated_entity
        raise KeyError(f"Can not update {self.__entity_name__} -> no id found")

    def create_entity(self, enable_logging: bool = False) -> dict:
        """Creates a new entity in Weclapp with the attributes set in the instance.
        Returns:
            dict: The created entity as a dictionary.
        """

        body = self.build_update_dictionary(update_type="full", creation_mode=True)
        if enable_logging:
            logging.warning(
                "Posting new %s to Weclapp with body: %s", type(self).__name__, body
            )
        entity_creation = Weclapp().post(entity_name=self.__entity_name__, body=body)

        return entity_creation

    @classmethod
    def from_blank(cls):
        """Creates a new empty instance of the class with all attributes set to
        their default values. This is useful for creating new entities without
        having to specify all attributes manually.
        """
        blank_entity = {}
        nullified_attributes = []

        for attribute, field_info in cls.model_fields.items():

            if getattr(field_info.annotation, "__origin__", None) in [List, list]:
                pass

            # Handle attributes that are weclapp subclasses (e.g. SalesOrderItems)
            elif isinstance(field_info.annotation, type) and issubclass(
                field_info.annotation, Blueprint
            ):
                blank_entity[attribute] = field_info.annotation.from_blank()

            # Handle regular attributes
            elif getattr(field_info.annotation, "__origin__", None) in [
                Optional,
                Union,
            ]:
                blank_entity[attribute] = None
                nullified_attributes.append(attribute)

            else:
                raise ValueError(
                    f"Unexpected attribute: {attribute} - {field_info.annotation}"
                )

        new_entity = cls(**blank_entity)

        for attribute in nullified_attributes:
            setattr(new_entity, attribute, None)

        new_entity.used_attributes = {}
        return new_entity

    def _handle_custom_attributes(
        self, value: list, update_type: Literal["full", "used"] = "used"
    ) -> List[dict]:
        """Handles the update of custom attributes. Iterates through the list of
        custom attributes, and adds their update dictionaries to the list of
        custom attributes if they were changed, or if update_type is "full".
        Args:
            value (list): List of custom attributes. Can be a list of
                WeclappMetaData instances or dictionaries.
            update_type (Literal["full", "used"]): Type of update.
        Returns:
            List[dict]: List of dictionaries representing the updated custom attributes.
        """
        updated_custom_attributes = []
        for custom_attribute in value:
            if isinstance(custom_attribute, WeclappMetaData):
                cat = custom_attribute.build_update_dictionary(update_type=update_type)
                if cat:
                    updated_custom_attributes.append(cat)
                elif update_type == "full":
                    updated_custom_attributes.append(cat)
            elif isinstance(custom_attribute, dict) and hasattr(
                custom_attribute, "attributeDefinitionId"
            ):
                updated_custom_attributes.append(custom_attribute)
        return updated_custom_attributes

    def _handle_list_values(
        self, key: str, value: Any, update_settings: UpdateSettings
    ) -> Any:
        """Handles the update of list values. If the list contains objects that
        are subclasses of Blueprint or dictionaries, it will be processed by
        _handle_list_of_objects method. If the list contains other types of
        values (e.g. int, str), it will return them as is.
        """
        if self.__annotations__.get(key, list) == list:
            if (
                len(value) > 0
                or update_settings.update_type == "full"
                or key in self.used_attributes
            ):
                return value

        else:
            items = self._handle_list_of_objects(value, update_settings=update_settings)
            if items:
                return items
        return None

    def _handle_list_of_objects(
        self, value: list, update_settings: UpdateSettings
    ) -> List[dict]:
        """Handles the update of list of dictionaries or objects. If the list
        contains objects that are subclasses of Blueprint, it will call their
        build_update_dictionary method. If the list contains dictionaries, it
        will return them as is. Versions of the objects will not be included to
        avoid optimistic lock errors.
        """
        items = []

        for item in value:
            if isinstance(item, Blueprint):
                if update_settings.creation_mode is False:
                    update_settings.update_type = "full"
                item_dict = item.build_update_dictionary(
                    update_type=update_settings.update_type,
                    creation_mode=update_settings.creation_mode,
                )
                if item_dict:
                    items.append(item_dict)
            elif isinstance(item, dict):
                items.append(item)

        return items

    def __setattr__(self, __name: str, __value: Any) -> None:
        """Sets an attribute of the instance.
        Args:
            __name (str): The name of the attribute to set.
            __value (Any): The value to set the attribute to.
        """

        if type(__value) in [int, str, float, bool]:
            target_type = self.__annotations__.get(__name, type(__value))
            args = get_args(target_type)

            # Converting value if necessary
            if int in args:
                if not isinstance(__value, int):
                    __value = int(float(__value))
            elif str in args:
                if not isinstance(__value, str):
                    __value = str(__value)
            elif bool in args:
                if not isinstance(__value, bool):
                    __value = bool(__value)
            elif float in args:
                if not isinstance(__value, float):
                    __value = float(__value)
            else:
                raise TypeError(
                    f"You tried to assign {type(__value).__name__} to "
                    f"the {target_type.__name__} attribute {__name} -> could "
                    f"not correct it."
                )

        if __value != getattr(self, __name):
            if __name not in ["used_attributes"]:
                self.used_attributes[__name] = getattr(self, __name)

            object.__setattr__(self, __name, __value)

    @property
    def __entity_name__(self) -> str:
        class_type = type(self)
        entity_name = str(class_type.__name__)
        entity_name = entity_name[:1].lower() + entity_name[1:]
        return entity_name



"""CustomAttributeClass for handling custom attributes in Weclapp."""


class CustomAttributeClass:
    """Class representing a custom attribute in Weclapp."""
    def __init__(
        self,
        group_name: str,
        attr_key: str,
        attr_id: str,
        attr_type: str,
        selectable_values: dict,
    ):
        self.group_name = group_name
        self.attr_key = attr_key
        self.attr_id = attr_id
        self.attr_type = attr_type
        self.selectable_values: dict = selectable_values

    def generate_python_line(self) -> str:
        """Generates a Python line for the custom attribute. This is used when
        creating the custom attributes class in Python.
        """
        return f"        self.{self.attr_key}: SimpleNamespace  # {self.attr_id}\n"

    def to_dict(self) -> dict:
        """Converts the custom attribute to a dictionary representation. This
        is used when creating named tuples or other data structures.
        """
        new = self.selectable_values.copy()
        new["id"] = self.attr_id
        new["attr_key"] = self.attr_key
        return new



"""This mopdule contains the Document class, which represents a document in
Weclapp and provides methods to interact with it."""

import io
from typing import Any, Union
from pydantic import BaseModel
from ..weclapp import Weclapp


class Document(BaseModel):
    """A class representing a document in Weclapp."""

    id: Union[str, None] = None
    version: Union[str, None] = None
    createdDate: Union[int, None] = None
    description: Union[str, None] = None
    lastModifiedDate: Union[int, None] = None
    mediaType: Union[str, None] = None
    name: Union[str, None] = None
    userId: Union[str, None] = None
    versions: list = []

    def __init__(self, **data: Any):
        super().__init__(**data)

    @classmethod
    def from_weclapp(cls, document_id: str = None, json_document: dict = None):
        """Creates a Document instance using either data from Weclapp entity or
        a provided JSON document."""
        if document_id is None and json_document is None:
            raise ValueError("Either document_id or json_document must be provided.")
        if json_document is None:
            if not isinstance(document_id, str):
                raise TypeError("document_id must be a string.")
            if not document_id.isdigit():
                raise ValueError("document_id must be a numeric string.")
            json_document = Weclapp().get(entity_name="document", entity_id=document_id)
        return cls(**json_document)

    def download_doc(self) -> io.BytesIO:
        """Downloads the document content as a BytesIO object."""
        content = Weclapp().get(
            entity_name="document", entity_id=f"{self.id}/download", as_type=bytes
        )
        file = io.BytesIO()
        file.write(content)
        file.seek(0)
        return file
